# Config.pm - The Object Class that provides a Config Object
#$Id: Config.pm,v 1.9 2004/01/28 22:52:14 james Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2003 HLR http://www.pcxperience.org  All rights reserved.
# Created by James Pattie, 08/06/2003.

package Portal::Accounting::Config;
use strict;
use Portal::AppState;
use HTMLObject::Base;
use HTMLObject::Normal;
use Portal::Accounting::Base;
use Portal::Accounting::DB::Main;
use Portal::Accounting::DB::Ledger;
use Portal::Accounting::Data::Config;
use Portal::Data::Variables;
use Portal::Accounting::Objects::Menu;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::AppState Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '0.01';

# global variables.

# modify to represent the commands this state provides.
my %commands;
%commands = (
  display => "Config Menu",
  close => "Close Window",
  );

=head1 NAME

Config - Object used to build a Config Menu for the accounting package

=head1 SYNOPSIS

  use Portal::Accounting::Config;
  my $obj = Portal::Accounting::Config->new;
  if ($obj->didErrorOccur())
  {
    die $obj->errorMessage();
  }
  $obj->run(command => $command);

=head1 DESCRIPTION

Config is the Config Menu from which every other part of
the systems will be configured.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new()

 Creates a new instance of the Portal::Accounting::Config module.
 See Portal::AppState(3) for a listing of required arguments.

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );

  if ($self->didErrorOccur)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module
  $self->{portalVariables} = Portal::Data::Variables->new(langObj => $self->{langObj});
  if ($self->{portalVariables}->didErrorOccur)
  {
    $self->setError($self->{portalVariables}->errorMessage);
    return $self;
  }

  # do validation
  if (!$self->Portal::Accounting::Config::isValid)
  {
    # the error is set in the isValid() method.
    return $self;
  }

  # do anything else you might need to do.

  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

=item HTMLObject run(command)

    Runs the specified command and returns the resulting HTMLObject.

=cut

sub run
{
  my $self = shift;
  my %args = ( command => "display", @_ );
  my $command = $args{command};
  my $doc;
  my $phrases = $self->{langObj};

  if (!exists $commands{$command})
  {
    $self->prefixError( sprintf($phrases->map("commandNotKnown"), $command) );
    return undef;
  }

  if ($command eq "close")
  {
    $doc = $self->close;
  }

  if ($command eq "display")
  {
    $doc = $self->display(command => $command);
  }

  if ($self->didErrorOccur)
  {
    $self->prefixError();
    return undef;
  }

  if (not defined $doc)
  {
    $self->prefixError("Command: $command returned undefined document!<br>\n");
    return undef;
  }
  return $doc;
}

sub display
{
  my $self = shift;
  my %args = ( command => "display", @_ );
  my $command = $args{command};
  my $doc;
  my $siteDir = $self->{configObj}->{siteDir};
  my $phrases = $self->{langObj};
  my $currVersion  = $Portal::Accounting::VERSION;
  my $mainHeight = "600";
  my $mainWidth = "1024";
  my $message;
  my $ledgerOpen = (exists $self->{input}->{ledgerOpen} ? $self->{input}->{ledgerOpen} : 0);

  my $windowName = (exists $self->{input}->{windowName} ? $self->{input}->{windowName} : "");
  if ($windowName)
  {
    # see if the window is registered and register if not.
    if (!exists $self->{sessionObj}->{store}->{windows}->{$windowName})
    {
      $self->{sessionObj}->{store}->{windows}->{$windowName} = $windowName;
      $self->{sessionObj}->{store}->{changed} = 1;
    }
  }
  else
  {
    $self->setError("You did not specify a Window Name!<br>\n");
    return undef;
  }

  $doc = HTMLObject::Normal->new;
  if (not defined $doc || $doc->didErrorOccur)
  {
    $self->prefixError("Could not create HTMLObject<br>\n");
    return undef;
  }

  # include the form_methods.js file
  $doc = $self->{methods}->includeCoreJSFiles(doc => $doc);
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError($self->{methods}->errorMessage);
    return undef;
  }

  my $ledgerCode = (exists $self->{input}->{ledgerCode} ? $self->{input}->{ledgerCode} : "");
  my $ledgerCodeURI = $doc->encodeString(string => $ledgerCode);

  # get the users colorScheme to work with.
  my $colorScheme = $self->{applicationObj}->getCurrentColorScheme(userId => $self->{userObj}->{id});
  if ($self->{applicationObj}->didErrorOccur)
  {
    $self->prefixError($self->{applicationObj}->errorMessage);
    return undef;
  }
  if (!defined $colorScheme)
  {
    $self->setError("No valid colorScheme returned!<br>\n");
    return undef;
  }

  # create the needed css entries
  $doc = $self->{methods}->generateColorSchemeCSS(doc => $doc, colorScheme => $colorScheme, appName => "Accounting");
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError($self->{methods}->errorMessage);
    return undef;
  }

  my $ledgerDBObj = Portal::Accounting::DB::Ledger->new(dbHandle => $self->{dbHandle}, langObj => $self->{langObj});
  if ($ledgerDBObj->didErrorOccur)
  {
    $self->setError($self->errorMessage . $ledgerDBObj->errorMessage);
    return undef;
    $message .= $self->errorMessage .  $ledgerDBObj->errorMessage;
  }

  # get ledgers in the system.
  my @Ledgers = $ledgerDBObj->getLedgers;
  if ($ledgerDBObj->didErrorOccur)
  {
    $self->setError($self->errorMessage . $ledgerDBObj->errorMessage);
    return undef;
    $message .= $self->errorMessage .  $ledgerDBObj->errorMessage;
  }

  # get accounts in the system.
  my @Accounts = $ledgerDBObj->getListOfAccounts(location => "global");
  if ($ledgerDBObj->didErrorOccur)
  {
    $self->postfixError($self->ledgerDBObj->errorMessage);
    return undef;
  }

  # get account assignments for the selected ledger.
  my @AssignedAccounts = ();
  if ($ledgerCode)
  {
    @AssignedAccounts = $ledgerDBObj->getListOfAccounts(location => "local", ledgerCode => $ledgerCode);
    if ($ledgerDBObj->didErrorOccur)
    {
      $self->postfixError($self->ledgerDBObj->errorMessage);
      return undef;
    }
  }

  $doc->setTitle("XIWA - ".$phrases->map("Config_Menu"));

  my $url = $self->{methods}->createBaseURL(type => "App", linkType => "cgi", appConfigObj => $self->{sessionObj}->{store}->{companyAppObj});
  my $refreshUrl = $url . "?app=Accounting&state=Config&command=display&ledgerCode=$ledgerCodeURI";

  $doc->setFocus("javascript");
  $doc->print(<<"END_OF_CODE");
function getLedgerCode()
{
  return document.mainForm.ledgerCode.options[document.mainForm.ledgerCode.selectedIndex].value;
}
function getLedger()
{
  return document.mainForm.ledgerCode.options[document.mainForm.ledgerCode.selectedIndex].text;
}
function refreshWindow()
{
  var d = document.mainForm;
  d.command.value="display";

  // refresh our parent window (should be the main menu)
  if (window.opener.refreshWindow)
  {
    window.opener.refreshWindow();
  }

  // now refresh our window
  d.submit();
  // window.location.reload(true);
}

function selectLedger()
{
  var d = document.mainForm;
  if (d.ledgerCode.selectedIndex < 1 || d.ledgerCode.selectedIndex >= d.ledgerCode.options.length)
  {
    alert("You need to select a Ledger to work with!");
    return;
  }
  d.command.value="display";
  d.submit();
}
END_OF_CODE

    # create the help JS function.
    $doc = $self->{methods}->displayJSHelper(doc => $doc, type => "help", langObj => $self->{langObj});
    if ($self->{methods}->didErrorOccur)
    {
      $self->setError($self->{methods}->errorMessage);
      return undef;
    }

    my $menuObj = Portal::Accounting::Objects::Menu->new(methods => $self->{methods},
                  sessionObj => $self->{sessionObj}, langObj => $self->{langObj}, userObj => $self->{userObj},
                  input => $self->{input}, portalSession => $self->{portalSession});
    if ($menuObj->didErrorOccur)
    {
      $self->setError($menuObj->errorMessage);
      return undef;
    }

    my $menuString = $menuObj->print(location => "menubar", state => "Config", command => $command, doc => $doc);
    if ($menuObj->didErrorOccur)
    {
      $self->setError($menuObj->errorMessage);
      return undef;
    }

    my $configMenuPhrase = $phrases->map("Config_Menu");
    my $accountsPhrase = $phrases->map("Accounts");
    my $subAccountsPhrase = $phrases->map("SubAccounts");
    my $stocksPhrase = $phrases->map("Stocks");
    my $ledgerPhrase = $phrases->map("Ledger");
    my $selectPhrase = $phrases->map("Select");
    my $newPhrase = $phrases->map("New");
    my $editPhrase = $phrases->map("Edit");
    my $deletePhrase = $phrases->map("Delete");
    my $pleaseCreateLedgerPhrase = $phrases->map("PleaseCreateLedger");
    my $monitorAccountsPhrase = $phrases->map("Monitor") . " " . $phrases->map("Accounts");
    my $subEntitiesPhrase = $phrases->map("Sub_Entities");
    my $assignedAccountsPhrase = $phrases->map("AssignedAccounts");
    my $employeesPhrase = $phrases->map("Employees");
    my $selectAllPhrase = $phrases->map("SelectAll");
    my $unSelectAllPhrase = $phrases->map("UnSelectAll");
    my $linksPhrase = $phrases->map("links");

    my %indexes = ( "ledgers" => "0");
    my $count;  # used by the different loops to keep track of the item they are on.

    $count = 1;
    my $ledgerOptions = "";
    my $ledgerName = "";
    my $ledgerNameURI = "";
    foreach my $ledgerObj (@Ledgers)
    {
      if ($ledgerObj->isValid)
      {
        $ledgerOptions .= "                    <option value=\"$ledgerObj->{code}\">$ledgerObj->{name}</option>\n";
        if ($ledgerObj->{code} eq $ledgerCode)
        {
          $indexes{ledgers} = $count;
          $ledgerName = $ledgerObj->{name};
        }
      }
      else
      {
        $message .= $phrases->map("Ignoring_invalid") . " ledgerObject: $ledgerObj->{name}<br>\n";
        $message .= $ledgerObj->errorMessage;
      }
      $count++;
    }
    my $deleteLedgerPhrase = sprintf($phrases->map("DeleteLedger"), $ledgerName);
    $ledgerNameURI = $doc->encodeString(string => $ledgerName);

    my $accountOptions = "";
    foreach my $accountObj (@Accounts)
    {
      $accountOptions .= "                    <option value=\"$accountObj->{number}\">$accountObj->{name}</option>\n";
    }

    my $assignedAccountsOptions = "";
    if ($ledgerCode)
    {
      foreach my $accountObj (@AssignedAccounts)
      {
        $assignedAccountsOptions .= "                    <option value=\"$accountObj->{number}\">$accountObj->{name}</option>\n";
      }
    }

    my $monitorAccountsURL = "$url?app=Accounting&state=Config::Monitor&command=display&ledger=$ledgerNameURI&ledgerCode=$ledgerCodeURI";

    $doc->setFocus("javascript");
    $doc->print(<<"END_OF_CODE");

function verifyDelete()
{
  return confirm("$deleteLedgerPhrase");
}
END_OF_CODE

    $doc->setFocus("body");
    $doc->print(<<"END_OF_CODE");
<center>
$menuString
<h2>$configMenuPhrase - XIWA $currVersion</h2>
<form name="mainForm" method="post">
  <input type="hidden" name="app" value="Accounting">
  <input type="hidden" name="state" value="Config">
  <input type="hidden" name="command" value="">
  <input type="hidden" name="windowName" value="$windowName">
  <input type="hidden" name="ledgerOpen" value="$ledgerOpen">
<table width="100%" border="0" cellpadding="0" cellspacing="0">
  <tbody>
  <tr valign="top">
    <td>
      <table border="0" cellpadding="0" cellspacing="0" width="100%">
        <tr>
          <td width="34%">
            <table class="configTables">
              <tr class="configHeader">
                <td colspan="2" align="center"><b>$accountsPhrase</b></td>
              </tr>
              <tr>
                <td>
                  <select name="account" multiple="multiple" size="5">
                  $accountOptions
                  </select>
                </td>
                <td valign="top" class="configLinks">
                  new<br>
                  edit<br>
                  delete<br>
                  assign<br>
                  <br>
                  <a class="item" href="#" onClick="selectAll(document.mainForm.account); return false;">$selectAllPhrase</a><br>
                  <a class="item" href="#" onClick="unSelectAll(document.mainForm.account); return false;">$unSelectAllPhrase</a>
                </td>
              </tr>
            </table>
          </td>
          <td width="33%">
            <table class="configTables">
              <tr class="configHeader">
                <td colspan="2" align="center" class="configTables"><b>$subAccountsPhrase</b></td>
              </tr>
              <tr>
                <td>
                  <select name="subAccount" size="5">
                  </select>
                </td>
                <td valign="top" class="configLinks">
                  new<br>
                  edit<br>
                  delete
                </td>
              </tr>
            </table>
          </td>
          <td width="33%">
            <table class="configTables">
              <tr class="configHeader">
                <td colspan="2" align="center"><b>$stocksPhrase</b></td>
              </tr>
              <tr>
                <td>
                  <select name="stock" size="5">
                  </select>
                </td>
                <td valign="top" class="configLinks">
                  new<br>
                  edit<br>
                  delete
                </td>
              </tr>
            </table>
          </td>
        </tr>
      </table>
    </td>
  </tr>
  <tr>
    <td><hr noshade></td>
  </tr>
  <tr valign="top">
    <td>
      <table width="100%" border="0" cellpadding="0" cellspacing="0">
        <tr>
          <td width="50%" align="left">
            <table class="configTables">
              <tr>
                <td class="configHeader"><b>$ledgerPhrase</b>
END_OF_CODE
  if (scalar @Ledgers > 0)
  {
    if (!$ledgerOpen)
    {
      $doc->print("                  <select name=\"ledgerCode\">\n");
      $doc->print("                    <option value=\"------\">- Select Ledger -</option>\n");
      $doc->print($ledgerOptions);
      $doc->print("                  </select>\n");
      $doc->print(<<"END_OF_CODE");
                  <input type="button" name="selectLedgerB" value="$selectPhrase" onClick="selectLedger();">
                </td>
                <td class="configLinks">
                  <a class="item" href="#" onclick="javascript: { var windowName = getTimeStamp(); window.open('$url?app=Accounting&state=Ledger&command=createLedger&windowName='+ windowName , windowName , 'width=$mainWidth,height=$mainHeight,scrollbars,status,toolbar'); return false; }">$newPhrase $ledgerPhrase</a>
                  <br>
END_OF_CODE
    }
    else
    {
      $doc->print("<input type=\"hidden\" name=\"ledgerCode\" value=\"$ledgerCode\">");
      $doc->print("<span style=\"color: " . $Ledgers[$indexes{ledgers}-1]->{color} . ";\">" . $Ledgers[$indexes{ledgers}-1]->{name} . "</span></td>");
      $doc->print(<<"END_OF_CODE");
                <td class="configLinks">
END_OF_CODE
    }
    if ($indexes{ledgers} > 0)
    {
      $doc->print(<<"END_OF_CODE");
                  <a class="item" href="#" onclick="javascript: { var windowName = getTimeStamp(); window.open('$url?app=Accounting&state=Ledger&windowName='+windowName+'&command=editLedger&ledgerCode=$ledgerCodeURI&ledger=$ledgerNameURI&ledgerOpen=$ledgerOpen', windowName , 'width=$mainWidth,height=$mainHeight,scrollbars,status,toolbar'); return false; }">$editPhrase $ledgerName</a>
END_OF_CODE
      if (!$ledgerOpen)
      {
        $doc->print(<<"END_OF_CODE");
                  <br>
                  <a class="item" href="#" onclick="javascript: { if (verifyDelete()) { var windowName = getTimeStamp(); window.open('$url?app=Accounting&state=Ledger&windowName='+windowName+'&command=deleteLedger&ledgerCode=$ledgerCodeURI&ledger=$ledgerNameURI', windowName , 'width=$mainWidth,height=$mainHeight,scrollbars,status,toolbar'); return false; } }">$deletePhrase $ledgerName</a>
END_OF_CODE
      }
    }
  }
  else
  {
    $doc->print(<<"END_OF_CODE");
                <a class="item" href="#" onclick="javascript: { var windowName = getTimeStamp(); window.open('$url?app=Accounting&state=Ledger&command=createLedger&windowName='+ windowName , windowName , 'width=$mainWidth,height=$mainHeight,scrollbars,status,toolbar'); return false; }">$newPhrase $ledgerPhrase</a>
END_OF_CODE
  }
  $doc->print(<<"END_OF_CODE");
                </td>
              </tr>
            </table>
          </td>
          <td width="50%" align="left">
END_OF_CODE
  if (scalar @Ledgers > 0)
  {
    if ($indexes{ledgers} > 0)
    {
      $doc->print(<<"END_OF_CODE");
            <table class="configTables">
              <tr>
                <td class="configHeader">
                  $ledgerName $linksPhrase:<br>
                  <a class="item" href="#" onClick="javascript: { var windowName = getTimeStamp(); window.open('$monitorAccountsURL&windowName='+ windowName , windowName , 'width=$mainWidth,height=$mainHeight,scrollbars,status,toolbar'); return false; }" title="$monitorAccountsPhrase">$monitorAccountsPhrase</a>
                </td>
              </tr>
            </table>
END_OF_CODE
    }
    else
    {
      $doc->print("&nbsp;");
    }
  }
  else
  {
    $doc->print(<<"END_OF_CODE");
            $pleaseCreateLedgerPhrase
END_OF_CODE
  }
  $doc->print(<<"END_OF_CODE");
          </td>
        </tr>
      </table>
    </td>
  </tr>
  <tr>
    <td>&nbsp;</td>
  </tr>
END_OF_CODE
  if ($indexes{ledgers} > 0)
  {
    $doc->print(<<"END_OF_CODE");
  <tr valign="top">
    <td>
      <table border="0" cellpadding="0" cellspacing="0" width="100%">
        <tr>
          <td width="34%">
            <table class="configTables">
              <tr class="configHeader">
                <td colspan="2" align="center"><b>$subEntitiesPhrase</b></td>
              </tr>
              <tr>
                <td>
                  <select name="subEntity" size="5">
                  </select>
                </td>
                <td valign="top" class="configLinks">
                  new<br>
                  edit<br>
                  delete<br>
                </td>
              </tr>
            </table>
          </td>
          <td width="33%">
            <table class="configTables">
              <tr class="configHeader">
                <td colspan="2" align="center" class="configTables"><b>$assignedAccountsPhrase</b></td>
              </tr>
              <tr>
                <td>
                  <select name="assignedAccounts" multiple="multiple" size="5">
                  $assignedAccountsOptions
                  </select>
                </td>
                <td valign="top" class="configLinks">
                  unassign<br>
                  <br>
                  <a class="item" href="#" onClick="selectAll(document.mainForm.assignedAccounts); return false;">$selectAllPhrase</a><br>
                  <a class="item" href="#" onClick="unSelectAll(document.mainForm.assignedAccounts); return false;">$unSelectAllPhrase</a>
                </td>
              </tr>
            </table>
          </td>
          <td width="33%">
            <table class="configTables">
              <tr class="configHeader">
                <td colspan="2" align="center"><b>$employeesPhrase</b></td>
              </tr>
              <tr>
                <td>
                  <select name="employee" size="5">
                  </select>
                </td>
                <td valign="top" class="configLinks">
                  new<br>
                  edit<br>
                  delete
                </td>
              </tr>
            </table>
          </td>
        </tr>
      </table>
    </td>
  </tr>
END_OF_CODE
  }

  $doc->print(<<"END_OF_CODE");
  </tbody>
</table>
</form>
</center>
END_OF_CODE

  $doc->print("<br><span style=\"color: red;\">$message</span>") if ($message);

  # setup our select box indexes.
  my $onloadCode = "";
  $onloadCode .= "document.mainForm.ledgerCode.selectedIndex=$indexes{ledgers};" if (!$ledgerOpen);
  $doc->setOnload(code => $onloadCode) if ($onloadCode);

  return $doc;
}

sub close
{
  my $self = shift;
  my %args = ( @_ );
  my $command = 'close';
  my $doc;
  my $phrases = $self->{langObj};

  $doc = HTMLObject::Normal->new;
  if (not defined $doc || $doc->didErrorOccur)
  {
    $self->setError("Could not create HTMLObject<br>\n");
    return undef;
  }

  # get the users colorScheme to work with.
  my $colorScheme = $self->{applicationObj}->getCurrentColorScheme(userId => $self->{userObj}->{id});
  if ($self->{applicationObj}->didErrorOccur)
  {
    $self->setError(errorString => $self->{applicationObj}->errorMessage);
    return undef;
  }
  if (!defined $colorScheme)
  {
    $self->setError(errorString => "No valid colorScheme returned!<br>\n");
    return undef;
  }

  # create the needed css entries
  $doc = $self->{methods}->generateColorSchemeCSS(doc => $doc, colorScheme => $colorScheme, appName => "Accounting");
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError(errorString => $self->{methods}->errorMessage);
    return undef;
  }

  my $window = $self->{input}->{windowName};
  if (exists $self->{sessionObj}->{store}->{windows}->{$window})
  {
    delete $self->{sessionObj}->{store}->{windows}->{$window};
    $self->{sessionObj}->{store}->{changed} = 1;
  }
  else
  {
    # this is an internal error!  I'm ignoring it for now.
  }

  # generate the document that will close itself.
  $doc->setTitle("XIWA - " . $phrases->map("Close_Window"));

  $doc->setOnload(code => "window.close();");  # output the JavaScript code that
                                              #will be run on the onLoad event of the page.

  return $doc;
}

=back

=cut

1;
__END__

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:xiwa@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::AppState(3)

=cut
