# Portal::Accounting::DB::Ledger.pm
# Created by JT Moree, Xperience, Inc.
# Thu Jan  9 13:45:24 UTC 2003

#$Id: Ledger.pm,v 1.21 2004/08/13 22:05:39 moreejt Exp $
#This code is licensed under the GPL

=head1 NAME

Ledger - Object used to interact with Database dealing with Ledger functions.

=head1 SYNOPSIS

  use Portal::Accounting::DB::Ledger;
  my $obj = Portal::Accounting::DB::Ledger->new(dbHandle => , langObj => );
  if ($obj->didErrorOccur())
  {
    die $obj->errorMessage();
  }

=head1 DESCRIPTION

Ledger contains methods for interchanging ledger data with the database.

=cut
package Portal::Accounting::DB::Ledger;
use strict;
use Portal::Accounting::Objects::Ledger;
use Portal::Accounting::DB::Main;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::DB::Main Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '.3';

=head1 Exported FUNCTIONS

=item scalar new()

    Creates a new instance of the Portal::Accounting::DB::Ledger
    object.

=cut
sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );

  if ($self->didErrorOccur)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module

  # do validation
  if (!$self->isValid)
  {
    # the error is set in the isValid() method.
    return $self;
  }

  # do anything else you might need to do.

  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (not $self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }
  # validate our parameters.

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

=item bool updateLedger(ledgerCode, color, businessType, cashOrAccrual)

    Updates a database ledger entry when given valid data.
    Only certain items can be changed on a ledger:
      color, businessType, cashOrAccrual

=cut
sub updateLedger
{
  my $self = shift;
  my %args = ( ledgerCode => "", color => "", businessType => "", cashOrAccrual => "", @_ ) ;
  my $color = "";
  my $businessType = "";
  my $cashOrAccrual = "";
  my $sql = "update ledger_tb set ";

  if (defined $args{ledgerObject})
  {
    if (! $args{ledgerObject}->isValid)
    {
      $self->setError("LedgerObject is invalid.<br>\n".$args{ledgerObject}->errorMessage);
      return 0;
    }
    $color = $args{ledgerObject}->{color};
    $cashOrAccrual = $args{ledgerObject}->{cashOrAccrual};
    $businessType = $args{ledgerObject}->{businessType};
  }
  elsif ($args{ledgerCode} ne "")
  {
    $color = $args{color};
    $cashOrAccrual = $args{cashOrAccrual};
    $businessType = $args{businessType};
  }
  else
  {
    $self->setError("LedgerCode or LedgerObject is required.<br>\n");
    return 0;
  }
  my $ledgerOrig = ($self->getLedgers(ledgerCode => $args{ledgerCode}))[0];
  if ($self->didErrorOccur)
  {
    $self->setError("Error while retrieving original LedgerObject for $args{ledgerCode}.<br>\n");
    return 0;
  }
  if ($color eq $ledgerOrig->{color})
  {  $color = ""; }
  if ($cashOrAccrual eq $ledgerOrig->{cashOrAccrual})
  {  $cashOrAccrual = ""; }
  if ($businessType eq $ledgerOrig->{businessType})
  {  $businessType = ""; }
  #must have 1 of three options
  my $num = 0;
  if ($args{color} ne "")
  {
    $num++;
    $sql .= " color = '$args{color}' ";
  }
  if ($num > 0)
  { $sql .= ", "; }
  if ($args{cashOrAccrual} ne "")
  {
    $num++;
    if ($args{cashOrAccrual} !~ /^(cash|accrual)$/i)
    {
      $self->setError("Invalid CashOrAccrual: '$args{cashOrAccrual}'.<br>\n");
      return 0;
    }
    my $cashOrAccrual = "";
    if ($args{cashOrAccrual} =~ /^cash$/i)
    {  $cashOrAccrual = "true";}
    elsif ($args{cashOrAccrual} =~ /^accrual$/i)
    { $cashOrAccrual = "false";}

    $sql .= " cash_or_accrual = '$cashOrAccrual'";
  }
  if ($num > 0)
  { $sql .= ", "; }
  if ($args{businessType} ne "")
  {
    $num++;
    $sql .= " business_type = '$args{businessType}' ";
  }
  if ($num < 1)
  {
    $self->setError("at least one field is required (color, cashOrAccrual, businessType).<br>\n");
    return 0;
  }
  $sql .= "WHERE code = '$args{ledgerCode}'";
  $self->{dbHandle}->write(sql => $sql);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError($self->{dbHandle}->errorMessage);
    return 0;
  }

  $self->{dbHandle}->commit;
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError($self->{dbHandle}->errorMessage);
    return 0;
  }

  return 1;
}

=item bool createLedger(ledger => Portal::Accounting::Objects::Ledger)

    Creates a database ledger entry when given a valid Ledger Object.

=cut
sub createLedger
{
  my $self = shift;
  my %args = ( ledger => undef, @_ ) ;
  my $cashOrAccrual = "";
  my $l = $args{ledger};
  my $ledgerName = "";

  if (not defined $args{ledger})
  {
    $self->setError("Error! Ledger not defined.  Cannot create!<br>\n");
    return 0;
  }
  if ($args{ledger}->isValid == 0)
  {
    $self->setError("Error! Ledger is invalid.  Cannot create!<br>\n");
    return 0;
  }
  if ($l->{cashOrAccrual} eq "cash")
  {  $cashOrAccrual = "true"; }
  else
  {  $cashOrAccrual = "false"; }
  #need to calculate start and end Dates for quarters for first year
  #technically this shoudl be done for each known year found in $l but there should only be one for creation
  my $fiscalYear = (sort keys %{$l->{knownYears}})[0];
  my @end = $l->getQuarterEnds($fiscalYear);
  if (not defined @end)
  {
    $self->setError("Could not insert '".$l->{name}."'! Bad Quarter End Calculations!<br>'\n" );
    return 0;
  }
  my @start = $l->getQuarterStarts($fiscalYear);
  if (not defined @start)
  {
    $self->setError("Could not insert '".$l->{name}."'! Bad Quarter Start Calculations!<br>'\n" );
    return 0;
  }
  my ($fyYear, $fyMo, $fyDay) = split ("-",$end[3]);
  my $insert = "insert into ledger_tb (name, code, color, cash_or_accrual, ";
  $insert .= "  fy_mo, fy_day, business_type," ;
  $insert .= "  start1, start2, start3, start4, start5, start6, start7, start8, start9, start10, start11, start12,";
  $insert .= "  offset1, offset2, offset3, offset4, offset5, offset6, offset7, offset8, offset9, offset10, offset11, offset12)";
  $insert .= " values ('$l->{name}', '$l->{code}', '$l->{color}', $cashOrAccrual, ";
  $insert .= "  '$fyMo', '$fyDay', $l->{businessType}, ";
  $insert .= " '".$l->start(1)."',";
  $insert .= " '".$l->start(2)."',";
  $insert .= " '".$l->start(3)."',";
  $insert .= " '".$l->start(4)."',";
  $insert .= " '".$l->start(5)."',";
  $insert .= " '".$l->start(6)."',";
  $insert .= " '".$l->start(7)."',";
  $insert .= " '".$l->start(8)."',";
  $insert .= " '".$l->start(9)."',";
  $insert .= " '".$l->start(10)."',";
  $insert .= " '".$l->start(11)."',";
  $insert .= " '".$l->start(12)."',";
  $insert .= " '".$l->offset(1)."',";
  $insert .= " '".$l->offset(2)."',";
  $insert .= " '".$l->offset(3)."',";
  $insert .= " '".$l->offset(4)."',";
  $insert .= " '".$l->offset(5)."',";
  $insert .= " '".$l->offset(6)."',";
  $insert .= " '".$l->offset(7)."',";
  $insert .= " '".$l->offset(8)."',";
  $insert .= " '".$l->offset(9)."',";
  $insert .= " '".$l->offset(10)."',";
  $insert .= " '".$l->offset(11)."',";
  $insert .= " '".$l->offset(12)."')";
  if ($l->didErrorOccur)
  {
    $self->prefixError($l->errorMessage."\n" );
    $self->prefixError("DEBUG: $insert<br>".$l->start(1)."\n" );
    return 0;
  }
  $self->{dbHandle}->write(sql => $insert);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError("Could not insert '$l->{name}'<br>\n" . $self->{dbHandle}->errorMessage);
    return 0;
  }
  $insert = "insert into known_years_tb ";
  $insert .= "(ledger_code, closed, year, ";
  $insert .= "start_date, end_date, q1_start, q1_end, q2_start, q2_end, q3_start, q3_end, q4_start, q4_end) ";
  $insert .= "values ('$l->{code}', 'false', '$fiscalYear', ";
  $insert .= "'$start[0]', '$end[3]', ";
  $insert .= "'$start[0]', '$end[0]', ";
  $insert .= "'$start[1]', '$end[1]', ";
  $insert .= "'$start[2]', '$end[2]', ";
  $insert .= "'$start[3]', '$end[3]') ";
  $self->{dbHandle}->write(sql => $insert);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError($self->{dbHandle}->errorMessage);
    return 0;
  }
  #############################
  #!!!!need to fill in variables_tb and entity_tb for CGI
  my $variables = "insert into variables_tb (working_year, ledger_code) values ('$fiscalYear', '$l->{code}')";
  my $entity = "insert into entity_tb (name, database_name) values ('$l->{name}', '$l->{name}')";
  $self->{dbHandle}->write(sql => $variables);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError($self->{dbHandle}->errorMessage);
    return 0;
  }
  $self->{dbHandle}->write(sql => $entity);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError($self->{dbHandle}->errorMessage);
    return 0;
  }
  ########################

  $self->{dbHandle}->commit;
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError($self->{dbHandle}->errorMessage);
    return 0;
  }

  return 1;
}

=item int deleteLedger(ledgerCode)

 Attempts to delete the specified ledger.
 Returns 1 on success, 0 on errors and sets the errorMessage,
 -1 if the ledger doesn't exist anymore.

=cut

sub deleteLedger
{
  my $self = shift;
  my %args = ( ledgerCode => "", @_ );
  my $ledgerCode = $args{ledgerCode};

  # first verify the ledger still exists to delete.
  my @ledger = $self->getLedgers(ledgerCode => $ledgerCode);
  if (@ledger)
  {
    my $ledger = $ledger[0];
    # first we need to remove the ledger name from the entity_tb table.
    $self->{dbHandle}->write(sql => "DELETE FROM entity_tb WHERE name = ?", plug => [ $ledger->{name} ]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # now we need to remove all entries from the time_tracker_config_tb and tips_tb tables that reference this ledger.
    $self->{dbHandle}->write(sql => "DELETE FROM time_tracker_config_tb WHERE ledger_code = ?", plug => [ $ledgerCode ]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    $self->{dbHandle}->write(sql => "DELETE FROM tips_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # ledger_tb
    $self->{dbHandle}->write(sql => "DELETE FROM ledger_tb WHERE code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # variables_tb
    $self->{dbHandle}->write(sql => "DELETE FROM variables_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # known_years_tb
    $self->{dbHandle}->write(sql => "DELETE FROM known_years_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # sub_entity_tb
    $self->{dbHandle}->write(sql => "DELETE FROM sub_entity_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # journal_entry_tb
    $self->{dbHandle}->write(sql => "DELETE FROM journal_entry_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # debitcredit_tb
    $self->{dbHandle}->write(sql => "DELETE FROM debitcredit_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # invoice_tb
    $self->{dbHandle}->write(sql => "DELETE FROM invoice_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # invoice_items_tb
    $self->{dbHandle}->write(sql => "DELETE FROM invoice_items_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # billing_tb
    $self->{dbHandle}->write(sql => "DELETE FROM billing_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # group_tb
    $self->{dbHandle}->write(sql => "DELETE FROM group_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # client_group_tb
    $self->{dbHandle}->write(sql => "DELETE FROM client_group_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # asset_tb
    $self->{dbHandle}->write(sql => "DELETE FROM asset_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # reminder_tb
    $self->{dbHandle}->write(sql => "DELETE FROM reminder_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # custom_report_tb
    $self->{dbHandle}->write(sql => "DELETE FROM custom_report_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # acct_bal_tb
    $self->{dbHandle}->write(sql => "DELETE FROM acct_bal_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # beginning_balance_entry_tb
    $self->{dbHandle}->write(sql => "DELETE FROM beginning_balance_entry_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # local_accounts_tb
    $self->{dbHandle}->write(sql => "DELETE FROM local_accounts_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # sec_inv_tb
    $self->{dbHandle}->write(sql => "DELETE FROM sec_inv_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # sec_event_tb
    $self->{dbHandle}->write(sql => "DELETE FROM sec_event_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # dep_sched_tb
    $self->{dbHandle}->write(sql => "DELETE FROM dep_sched_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # report_index_tb
    $self->{dbHandle}->write(sql => "DELETE FROM report_index_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # report_tb
    $self->{dbHandle}->write(sql => "DELETE FROM report_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # memorized_report_tb
    $self->{dbHandle}->write(sql => "DELETE FROM memorized_report_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # tax_loss_tb
    $self->{dbHandle}->write(sql => "DELETE FROM tax_loss_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # scratch_tb
    $self->{dbHandle}->write(sql => "DELETE FROM scratch_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    # lot_node_tb
    $self->{dbHandle}->write(sql => "DELETE FROM lot_node_tb WHERE ledger_code = ?", plug => [ $ledgerCode]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }

    $self->{dbHandle}->commit;
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError($self->{dbHandle}->errorMessage);
      return 0;
    }
  }
  else
  {
    return -1;
  }

  return 1;
}

=item int closeoutLedger(ledgerCode, date, equityAccounts)

 Closes out ledger on the given date using the supplied equity account(s) to balance the transactions.
 Returns 1 on success, 0 on errors and sets the errorMessage.  commit is set to 1 by default.  This
 will commit all work done to the database at the end of the function by default.  To override, set
 commit to 0.  You will then need to commit the transaction(s) to the database.

=cut

sub closeoutLedger
{
  my $self = shift;
  my %args = ( ledgerCode => "", closeoutDate => "", equityAccounts => undef, leftoverEquityAccount => "", percentages => undef, checkDateRange => 1, commit => 1, @_ );
  my $ledgerCode = $args{ledgerCode};
  my $closeoutDate = $args{closeoutDate};
  my $equityAccounts = $args{equityAccounts};
  my $leftoverEquityAccount = $args{leftoverEquityAccount};
  my $checkDateRange = $args{checkDateRange};
  my $fromDate;
  my $toDate;
  my $workingYearEnd;
  my $sth;
  
#!!! get rid of this once sessions store by ledgerCode
  $sth = $self->{dbHandle}->read(sql => "SELECT name FROM ledger_tb WHERE code = '$ledgerCode'");
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError("Could not lookup name for ledgerCode: $ledgerCode");
    return 0;
  }
  my @ledger = $sth->fetchrow_array;
  if (not defined $ledger[0])
  {
    $self->setError("Could not lookup name for ledgerCode: $ledgerCode");
    return 0;
  }
  my $ledgerName = $ledger[0];
#!!!!
  
  # validate input
  if (not defined $ledgerCode || $ledgerCode eq "")
  {
    $self->setError(errorString => sprintf($self->{langObj}->map("invalidArgument"), "ledgerCode", $ledgerCode));
    return 0;
  }
  if (not defined $closeoutDate || $closeoutDate eq "")
  {
    $self->setError(errorString => sprintf($self->{langObj}->map("invalidArgument"), "closeoutDate", $closeoutDate));
    return 0;
  }
  if (not defined $equityAccounts)
  {
    $self->setError(errorString => $self->{langObj}->map("equityAccountsArrayReferenceRequired"));
    return 0;
  }

  # convert equityAccounts into array
  my @eqAccounts = @{$equityAccounts};

  # get the current fiscal year
  my $currentFiscalYear = $self->getCurrentFiscalYear($ledgerCode);
  $fromDate = $self->{sessionObj}->{store}->{ledgers}->{$ledgerName}->start(month => 1, fiscalYear => $currentFiscalYear);
  $toDate = $self->{sessionObj}->{store}->{ledgers}->{$ledgerName}->end(month => 12, fiscalYear => $currentFiscalYear);
  $workingYearEnd = $toDate;

  if ($closeoutDate =~ /(-|\/)/)
  {
    my ($result, $fixedCloseoutDate, $errorString);
    ($result, $fixedCloseoutDate, $errorString) = $self->{methods}->fixupDate(date => $closeoutDate, seperator => '-');
    if ($result == 0)
    {
      $self->setError(errorString => sprintf($self->{langObj}->map("errorFixingUpDate"), $closeoutDate) . "\n" . $errorString);
      return 0;
    }
    $closeoutDate = $fixedCloseoutDate;  # assign the good date to the original variable
    $toDate = $closeoutDate;

    # get the earliest closeout date (if any)
    my @previousCloseoutDates = $self->getCurrentFiscalCloseoutDates(ledgerCode => $ledgerCode);
    if ($self->didErrorOccur)
    {
      return 0;
    }

    if (scalar @previousCloseoutDates > 0)
    {
      $fromDate = $previousCloseoutDates[0];
    }
  }
  elsif ($closeoutDate =~ /^\d+$/)
  {
    # closeout the current fiscal year by setting $closeoutDate to the last day of the fiscal year to closeout
    $closeoutDate = $workingYearEnd;
  }
  else
  {
    $self->setError(errorString => sprintf($self->{langObj}->map("invalidArgument"), "closeoutDate", $closeoutDate));
    return 0;
  }

  my ($result, $fixedDate, $errorString);
  ($result, $fixedDate, $errorString) = $self->{methods}->fixupDate(date => $closeoutDate, seperator => "-");
  if ($result == 0)
  {
    $self->setError(errorString => sprintf($self->{langObj}->map("errorFixingUpDate"), $closeoutDate) . "<br>\n" . $errorString);
    return 0;
  }
  $closeoutDate = $fixedDate;

  ($result, $fixedDate, $errorString) = $self->{methods}->fixupDate(date => $fromDate, seperator => "-");
  if ($result == 0)
  {
    $self->setError(errorString => sprintf($self->{langObj}->map("errorFixingUpDate"), $fromDate) . "<br>\n" . $errorString);
    return 0;
  }
  $fromDate = $fixedDate;

  ($result, $fixedDate, $errorString) = $self->{methods}->fixupDate(date => $toDate, seperator => "-");
  if ($result == 0)
  {
    $self->setError(errorString => sprintf($self->{langObj}->map("errorFixingUpDate"), $toDate) . "<br>\n" . $errorString);
    return 0;
  }
  $toDate = $fixedDate;

  ($result, $fixedDate, $errorString) = $self->{methods}->fixupDate(date => $workingYearEnd, seperator => "-");
  if ($result == 0)
  {
    $self->setError(errorString => sprintf($self->{langObj}->map("errorFixingUpDate"), $workingYearEnd) . "<br>\n" . $errorString);
    return 0;
  }
  $workingYearEnd = $fixedDate;

  if ($checkDateRange ne "0")
  {
    # make sure closeoutDate is validly between the previous closeout date
    if (($fromDate gt $closeoutDate) || ($toDate lt $closeoutDate))
    {
      $self->setError(errorString => sprintf($self->{langObj}->map("closeoutDateInvalid"), $closeoutDate, $fromDate));
      return 0;
    }
  }

  # verify that the closeoutDate has not already been used or that the ledger has not been closed out after closeoutDate
  $sth = $self->{dbHandle}->read(sql => "SELECT DISTINCT(bal_date) FROM acct_bal_tb WHERE ledger_code = ? AND bal_date >= ?", plug => [ $ledgerCode, $closeoutDate ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => sprintf($self->{langObj}->map("errorLookingUpCloseouts"), $closeoutDate));
    return 0;
  }
  if (defined $sth->fetchrow_array)
  {
    $self->setError(errorString => sprintf($self->{langObj}->map("ledgerAlreadyClosed"), $closeoutDate));
    return 0;
  }

  # get business type
  my $ledgerObj = $self->getLedger(ledgerCode => $ledgerCode);
  if ((not defined $ledgerObj) || ($self->didErrorOccur) || (not $ledgerObj->isValid))
  {
    $self->setError(errorString => sprintf($self->{langObj}->map("invalidLedger"), $ledgerCode) . "<br>\n" . $self->errorMessage);
    return 0;
  }
  my $businessType = $ledgerObj->businessType;

  # get the last closeout period date
  my $lastCloseoutDate;
  $sth = $self->{dbHandle}->read(sql => "SELECT DISTINCT(bal_date) FROM acct_bal_tb WHERE ledger_code = ? ORDER BY bal_date DESC", plug => [ $ledgerCode ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => sprintf($self->{langObj}->map("errorGettingLastCloseoutPeriod"), $ledgerCode));
    return 0;
  }
  my @info = $sth->fetchrow_array;
  if (not defined $info[0])
  {
    $lastCloseoutDate = "Never";
  }
  else
  {
    $lastCloseoutDate = $info[0];
  }

  # validate the equity account(s) sent in for each business type
  if ($businessType == 1)
  {
    if (scalar @eqAccounts <= 1)
    {
      $self->setError(errorString => $self->{langObj}->map("chooseTwoOrMoreEquityAccounts"));
      return 0;
    }

    if (not defined $leftoverEquityAccount || length $leftoverEquityAccount == 0)
    {
      $self->setError(errorString => $self->{langObj}->map("chooseLeftoverEquityAccount"));
      return 0;
    }

    foreach my $equityAccount (@eqAccounts)
    {
      if (not $self->checkEquityAccount(accountNumber => $equityAccount, ledgerCode => $ledgerCode))
      {
        $self->setError(errorString => $self->errorMessage);
        return 0;
      }
    }

    if (not $self->checkEquityAccount(accountNumber => $leftoverEquityAccount, ledgerCode => $ledgerCode))
    {
      $self->setError(errorString => $self->errorMessage);
      return 0;
    }
  }
  elsif ($businessType == 3)
  {
    if (scalar @eqAccounts <= 1)
    {
      $self->setError(errorString => $self->{langObj}->map("chooseTwoOrMoreEquityAccounts"));
      return 0;
    }

    my @percentages = @{$args{percentages}};

    if (not defined @percentages || scalar @percentages == 0)
    {
      $self->setError(errorString => sprintf($self->{langObj}->map("invalidArgument"), "percentages", @percentages));
      return 0;
    }

    if (scalar @percentages != scalar @eqAccounts)
    {
      $self->setError(errorString => $self->{langObj}->map("notEnoughPercentages"));
      return 0;
    }

    if (not defined $leftoverEquityAccount || length $leftoverEquityAccount == 0)
    {
      $self->setError(errorString => $self->{langObj}->map("chooseLeftoverEquityAccount"));
      return 0;
    }

    my $totalPercentages = 0;
    for (my $count=0; $count < scalar @eqAccounts; $count++)
    {
      if (not $self->checkEquityAccount(accountNumber => $eqAccounts[$count], ledgerCode => $ledgerCode))
      {
        $self->setError(errorString => $self->errorMessage);
        return 0;
      }

      if ($percentages[$count] <= 0 || $percentages[$count] >= 100)
      {
        $self->setError(errorString => sprintf($self->{langObj}->map("percentageInvalid"), $percentages[$count], $eqAccounts[$count]));
      }

      $totalPercentages += $percentages[$count];
    }
    # verify percentages input add up to 100%
    if ($totalPercentages != 100)
    {
      $self->setError(errorString => sprintf($self->{langObj}->map("sumOfEquityAccountPercentagesNot100"), $totalPercentages));
    }

    if (not $self->checkEquityAccount(accountNumber => $leftoverEquityAccount, ledgerCode => $ledgerCode))
    {
      $self->setError(errorString => $self->errorMessage);
      return 0;
    }
  }
  else
  {
    if (scalar @eqAccounts != 1)
    {
      $self->setError(errorString => $self->{langObj}->map("chooseOneEquityAccount"));
      return 0;
    }

    if (not $self->checkEquityAccount(accountNumber => $eqAccounts[0], ledgerCode => $ledgerCode))
    {
      $self->setError(errorString => $self->errorMessage);
      return 0;
    }
  }

  # verify that debits == credits for all transactions in the system
  my $systemBalances = $self->doesSystemBalance(ledgerCode => $ledgerCode, beginDate => undef, endDate => $closeoutDate);
  if ($self->didErrorOccur || $systemBalances == 0)
  {
    return 0;
  }
  if ($systemBalances == -1)
  {
    $self->setError(errorString => $self->{langObj}->map("debitsGreaterThanCredits"));
    return 0;
  }
  elsif ($systemBalances == -2)
  {
    $self->setError(errorString => $self->{langObj}->map("creditsGreaterThanDebits"));
    return 0;
  }

  # Use the balance stored procedures for the date passed in and the ledger code to gather the balances, accounts, and sub-accounts
  # that are needed to close out that period.
  my $subAccountCloseoutSQL = "INSERT INTO acct_bal_tb (ledger_code, acct_num, balance, sub_acct_num, bal_date) \
  (SELECT ledger_code, account_num, \
  sub_account_balance_plus_activity(account_num, sub_account_num, ledger_code, '$closeoutDate', 'custom'), sub_account_num, '$closeoutDate'::date \
  FROM debitcredit_tb \
  WHERE ledger_code = ? AND \
  trans_date <= ? AND \
  account_num = (SELECT number FROM account_tb WHERE number = account_num AND \
  type = (SELECT num FROM acct_type_tb WHERE num = type AND num <> 0 AND num <> 17 AND first_digit <> '4' AND first_digit <> '5')) \
  UNION \
  SELECT ledger_code, acct_num, \
  sub_account_balance_plus_activity(acct_num, sub_acct_num, ledger_code, '$closeoutDate', 'custom'), sub_acct_num, '$closeoutDate' \
  FROM acct_bal_tb \
  WHERE ledger_code = ? AND \
  bal_date <= ? AND \
  acct_num = (SELECT number FROM account_tb WHERE number = acct_num AND \
  type = (SELECT num FROM acct_type_tb WHERE num = type AND num <> 0 AND num <> 10 AND num <> 17 AND first_digit <> '4' AND first_digit <> '5')) \
  AND sub_acct_num <> -2 \
  GROUP BY ledger_code, acct_num, sub_acct_num)";
  $sth = $self->{dbHandle}->write(sql => $subAccountCloseoutSQL, plug => [ $ledgerCode, $closeoutDate, $ledgerCode, $closeoutDate ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => $self->{langObj}->map("errorInsertingSubAccountBalances") . "<br>\n" . $self->{dbHandle}->errorMessage);
    return 0;
  }

  # Use the balance stored procedures for the date passed in and the ledger code to gather the balances and accounts
  # that are needed to close out that period.
  my $accountCloseoutSQL = "INSERT INTO acct_bal_tb (ledger_code, acct_num, balance, sub_acct_num, bal_date) \
  (SELECT ledger_code, account_num, \
  account_balance_plus_activity(account_num, ledger_code, '$closeoutDate', 'custom'), -2, '$closeoutDate'::date \
  FROM debitcredit_tb \
  WHERE ledger_code = ? AND \
  trans_date <= ? AND \
  account_num = (SELECT number FROM account_tb WHERE number = account_num AND \
  type = (SELECT num FROM acct_type_tb WHERE num = type AND num <> 17 AND first_digit <> '4' AND first_digit <> '5')) \
  GROUP BY ledger_code, account_num \
  UNION \
  SELECT ledger_code, acct_num, \
  account_balance_plus_activity(acct_num, ledger_code, '$closeoutDate', 'custom'), -2, '$closeoutDate' \
  FROM acct_bal_tb \
  WHERE ledger_code = ? AND \
  bal_date <= ? AND \
  acct_num = (SELECT number FROM account_tb WHERE number = acct_num AND \
  type = (SELECT num FROM acct_type_tb WHERE num = type AND num <> 17 AND first_digit <> '4' AND first_digit <> '5')) \
  GROUP BY ledger_code, acct_num)";
  $sth = $self->{dbHandle}->write(sql => $accountCloseoutSQL, plug => [ $ledgerCode, $closeoutDate, $ledgerCode, $closeoutDate ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => $self->{langObj}->map("errorInsertingAccountBalances") . "<br>\n" . $self->{dbHandle}->errorMessage);
    return 0;
  }

  # retrieve the last previous balance date
  $sth = $self->{dbHandle}->read(sql => "SELECT MAX(bal_date) FROM acct_bal_tb WHERE ledger_code = ? AND bal_date < ?", plug => [ $ledgerCode, $closeoutDate ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => $self->{langObj}->map("errorLookingUpPreviousCloseoutDate") . "<br>\n" . $self->{dbHandle}->errorMessage);
    return 0;
  }
  my $lastPreviousCloseoutDate = "";
  if (defined (@info = $sth->fetchrow_array) && defined $info[0])
  {
    $lastPreviousCloseoutDate = $info[0];
  }
  # gather activity for Revenues and Expenses and Dividends Paid (account type == 17)
  my $dividendsPaidActivitySQL = "SELECT activity_sum(true, (SELECT SUM(amount) FROM debitcredit_tb WHERE ledger_code = ? AND account_num = (SELECT number FROM account_tb WHERE number = account_num AND type = 17) ##?## AND trans_date <= ? AND dorc = true), (SELECT SUM(amount) FROM debitcredit_tb WHERE ledger_code = ? AND account_num = (SELECT number FROM account_tb WHERE number = account_num AND type = 17) ##?## AND trans_date <= ? AND dorc = false))";
  if ($lastPreviousCloseoutDate eq "")
  {
    $dividendsPaidActivitySQL =~ s/\#\#\?\#\#//g;
  }
  else
  {
    $dividendsPaidActivitySQL =~ s/\#\#\?\#\#/AND trans_date > '$lastPreviousCloseoutDate'/g;
  }
  $sth = $self->{dbHandle}->read(sql => $dividendsPaidActivitySQL, plug => [ $ledgerCode, $closeoutDate, $ledgerCode, $closeoutDate ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => $self->{langObj}->map("errorRetrievingActivityForDividendsPaid") . "<br>\n" . $self->{dbHandle}->errorMessage);
    return 0;
  }
  my $dividendsPaidTotal = "0.00";
  if (defined (@info = $sth->fetchrow_array) && defined $info[0])
  {
    $dividendsPaidTotal = $info[0];
  }

  my $revenuesActivitySQL = "SELECT activity_sum(false, (SELECT SUM(amount) FROM debitcredit_tb WHERE ledger_code = ? AND account_num = (SELECT number FROM account_tb WHERE number = account_num AND type = (SELECT num FROM acct_type_tb WHERE num = type AND first_digit = '4')) ##?## AND trans_date <= ? AND dorc = true), (SELECT SUM(amount) FROM debitcredit_tb WHERE ledger_code = ? AND account_num = (SELECT number FROM account_tb WHERE number = account_num AND type = (SELECT num FROM acct_type_tb WHERE num = type AND first_digit = '4')) ##?## AND trans_date <= ? AND dorc = false))";
  if ($lastPreviousCloseoutDate eq "")
  {
    $revenuesActivitySQL =~ s/\#\#\?\#\#//g;
  }
  else
  {
    $revenuesActivitySQL =~ s/\#\#\?\#\#/AND trans_date > '$lastPreviousCloseoutDate'/g;
  }
  $sth = $self->{dbHandle}->read(sql => $revenuesActivitySQL, plug => [ $ledgerCode, $closeoutDate, $ledgerCode, $closeoutDate ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => $self->{langObj}->map("errorRetrievingActivityForRevenues") . "<br>\n" . $self->{dbHandle}->errorMessage);
    return 0;
  }
  my $revenuesTotal = "0.00";
  if (defined (@info = $sth->fetchrow_array) && defined $info[0])
  {
    $revenuesTotal = $info[0];
  }

  my $expensesActivitySQL = "SELECT activity_sum(true, (SELECT SUM(amount) FROM debitcredit_tb WHERE ledger_code = ? AND account_num = (SELECT number FROM account_tb WHERE number = account_num AND type = (SELECT num FROM acct_type_tb WHERE num = type AND first_digit = '5')) ##?## AND trans_date <= ? AND dorc = true), (SELECT SUM(amount) FROM debitcredit_tb WHERE ledger_code = ? AND account_num = (SELECT number FROM account_tb WHERE number = account_num AND type = (SELECT num FROM acct_type_tb WHERE num = type AND first_digit = '5')) ##?## AND trans_date <= ? AND dorc = false))";
  if ($lastPreviousCloseoutDate eq "")
  {
    $expensesActivitySQL =~ s/\#\#\?\#\#//g;
  }
  else
  {
    $expensesActivitySQL =~ s/\#\#\?\#\#/AND trans_date > '$lastPreviousCloseoutDate'/g;
  }
  $sth = $self->{dbHandle}->read(sql => $expensesActivitySQL, plug => [ $ledgerCode, $closeoutDate, $ledgerCode, $closeoutDate ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => $self->{langObj}->map("errorRetrievingActivityForExpenses") . "<br>\n" . $self->{dbHandle}->errorMessage);
    return 0;
  }
  my $expensesTotal = "0.00";
  if (defined (@info = $sth->fetchrow_array) && defined $info[0])
  {
    $expensesTotal = $info[0];
  }

  # format activity totals
  $dividendsPaidTotal = sprintf("%.2f", $dividendsPaidTotal);
  $revenuesTotal = sprintf("%.2f", $revenuesTotal);
  $expensesTotal = sprintf("%.2f", $expensesTotal);

  # calculate the Retained Earnings balance
  my $retainedEarnings = $revenuesTotal - ($expensesTotal + $dividendsPaidTotal);
  $retainedEarnings = sprintf("%.2f", $retainedEarnings);

  # adjust the user-entered equity account(s) by the Retained Earnings amount
  if ($businessType == 1)
  {
    my $partnerTotal;
    my $partnerRemainder;
    if ($retainedEarnings ne "0.00")
    {
      $partnerTotal = $retainedEarnings / scalar @eqAccounts;  # calculate how much each partner receives
      if ($partnerTotal < 0)
      {
        $partnerTotal += "0.005";
      }
      else
      {
        $partnerTotal -= "0.005";
      }
      $partnerTotal = sprintf("%.2f", $partnerTotal);

      $partnerRemainder = scalar @eqAccounts * $partnerTotal;  # calculate how much total was divided up
      if ($partnerRemainder < 0)
      {
        $partnerRemainder += "0.005";
      }
      else
      {
        $partnerRemainder -= "0.005";
      }
      $partnerRemainder = sprintf("%.2f", $partnerRemainder);

      $partnerRemainder = $retainedEarnings - $partnerRemainder;  # calculate the remainder (if any)
      if ($partnerRemainder < 0)
      {
        $partnerRemainder = sprintf("%.2f", $partnerRemainder);
      }
    }
    else
    {
      $partnerTotal = $partnerRemainder = "0.00";  # there can be no division of 0.00!
    }

    for (my $count = 0; $count < scalar @eqAccounts; $count++)
    {
      $sth = $self->{dbHandle}->read(sql => "SELECT balance FROM acct_bal_tb WHERE acct_num = ? AND sub_acct_num = '-2' AND bal_date = ? AND ledger_code = ?", plug => [ $eqAccounts[$count], $toDate, $ledgerCode ]);
      if ($self->{dbHandle}->didErrorOccur)
      {
        $self->setError(errorString => sprintf($self->{langObj}->map("errorLookingUpBalanceForPartnerEquityAccount"),) . "<br>\n" . $self->{dbHandle}->errorMessage);
        return 0;
      }

      my $partnerBalanceAmount = "0.00";
      if (defined (@info = $sth->fetchrow_array) && defined $info[0])
      {
        # add to existing balance entries
        $partnerBalanceAmount = $info[0];
        $partnerBalanceAmount += $partnerTotal;
        $partnerBalanceAmount = sprintf("%.2f", $partnerBalanceAmount);

        $sth = $self->{dbHandle}->write(sql => "UPDATE acct_bal_tb SET balance = ? WHERE acct_num = ? AND sub_acct_num = '-2' AND bal_date = ? AND ledger_code = ?", plug => [ $partnerBalanceAmount, $eqAccounts[$count], $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorUpdatingRetainedEarningsBalanceForPartnerEquityAccount"), $eqAccounts[$count], $partnerBalanceAmount, $toDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }

        $sth = $self->{dbHandle}->read(sql => "SELECT balance FROM acct_bal_tb WHERE acct_num = ? AND sub_acct_num = '-1' AND bal_date = ? AND ledger_code = ?", plug => [ $eqAccounts[$count], $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorLookingUpBalanceForPartnerEquitySubAccount"), $eqAccounts[$count], $businessType) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }
        if (defined (@info = $sth->fetchrow_array))
        {
          $partnerBalanceAmount = $info[0];
          $partnerBalanceAmount += $partnerTotal;
          $partnerBalanceAmount = sprintf("%.2f", $partnerBalanceAmount);
        }
        else
        {
          $self->setError(errorString => $self->{langObj}->map("errorNoBalancedReturnedForPartnerEquitySubAccount"));
          return 0;
        }

        $sth = $self->{dbHandle}->write(sql => "UPDATE acct_bal_tb SET balance = ? WHERE acct_num = ? AND sub_acct_num = '-1' AND bal_date = ? AND ledger_code = ?", plug => [ $partnerBalanceAmount, $eqAccounts[$count], $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorUpdatingRetainedEarningsBalanceForPartnerEquitySubAccount"), $eqAccounts[$count], $partnerBalanceAmount, $toDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }
      }
      else
      {
        # create new balance entries
        $sth = $self->{dbHandle}->write(sql => "INSERT INTO acct_bal_tb (acct_num, sub_acct_num, balance, bal_date, ledger_code) VALUES (?, '-2', ?, ?, ?)", plug => [ $eqAccounts[$count], $partnerTotal, $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorCreatingRetainedEarningsBalanceForPartnerEquityAccount"),$eqAccounts[$count], $partnerTotal, $toDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }
        $sth = $self->{dbHandle}->write(sql => "INSERT INTO acct_bal_tb (acct_num, sub_acct_num, balance, bal_date, ledger_code) VALUES (?, '-1', ?, ?, ?)", plug => [ $eqAccounts[$count], $partnerTotal, $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorCreatingRetainedEarningsBalanceForPartnerEquitySubAccount"),$eqAccounts[$count], $partnerTotal, $toDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }
      }
    }

    if ($partnerRemainder ne "0.00")
    {
      # obtain the equity account balance, to add to the total it already has
      $sth = $self->{dbHandle}->read(sql => "SELECT balance FROM acct_bal_tb WHERE acct_num = ? AND sub_acct_num = '-2' AND bal_date = ? AND ledger_code = ?", plug => [ $leftoverEquityAccount, $toDate, $ledgerCode ]);
      if ($self->{dbHandle}->didErrorOccur)
      {
        $self->setError(errorString => sprintf($self->{langObj}->map("errorLookingUpBalanceForPartnerLeftoverEquityAccount"), $leftoverEquityAccount) . "<br>\n" . $self->{dbHandle}->errorMessage);
        return 0;
      }

      my $leftoverBalanceAmount = "0.00";
      if (defined (@info = $sth->fetchrow_array) && defined $info[0])
      {
        # add to existing balance entries
        $leftoverBalanceAmount = $info[0];
        $leftoverBalanceAmount += $partnerRemainder;
        $leftoverBalanceAmount = sprintf("%.2f", $leftoverBalanceAmount);

        $sth = $self->{dbHandle}->write(sql => "UPDATE acct_bal_tb SET balance = ? WHERE acct_num = ? AND sub_acct_num = '-2' AND bal_date = ? AND ledger_code = ?", plug => [ $leftoverBalanceAmount, $leftoverEquityAccount, $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorUpdatingRetainedEarningsBalanceForLeftoverEquityAccount"), $leftoverEquityAccount, $leftoverBalanceAmount, $toDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }

        $sth = $self->{dbHandle}->read(sql => "SELECT balance FROM acct_bal_tb WHERE acct_num = ? AND sub_acct_num = '-1' AND bal_date = ? AND ledger_code = ?", plug => [ $leftoverEquityAccount, $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorLookingUpBalanceForPartnerLeftoverEquitySubAccount"), $leftoverEquityAccount, $businessType) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }
        if (defined (@info = $sth->fetchrow_array) && defined $info[0])
        {
          $leftoverBalanceAmount = $info[0];
          $leftoverBalanceAmount += $partnerRemainder;
          $leftoverBalanceAmount = sprintf("%.2f", $leftoverBalanceAmount);
        }
        else
        {
          $self->setError(errorString => $self->{langObj}->map("errorNoBalancedReturnedForLeftoverEquitySubAccount"));
          return 0;
        }

        $sth = $self->{dbHandle}->write(sql => "UPDATE acct_bal_tb SET balance = ? WHERE acct_num = ? AND sub_acct_num = '-1' AND bal_date = ? AND ledger_code = ?", plug => [ $leftoverBalanceAmount, $leftoverEquityAccount, $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorUpdatingRetainedEarningsBalanceForLeftoverEquitySubAccount"), $leftoverEquityAccount, $leftoverBalanceAmount, $toDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }
      }
      else
      {
        # create new balance entries
        $sth = $self->{dbHandle}->write(sql => "INSERT INTO acct_bal_tb (acct_num, sub_acct_num, balance, bal_date, ledger_code) VALUES (?, '-2', ?, ?, ?)", plug => [ $leftoverEquityAccount, $partnerRemainder, $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorCreatingRetainedEarningsBalanceForLeftoverEquityAccount"), $leftoverEquityAccount, $partnerRemainder, $toDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }
        $sth = $self->{dbHandle}->write(sql => "INSERT INTO acct_bal_tb (acct_num, sub_acct_num, balance, bal_date, ledger_code) VALUES (?, '-1', ?, ?, ?)", plug => [ $leftoverEquityAccount, $partnerRemainder, $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorCreatingRetainedEarningsBalanceForLeftoverEquitySubAccount"), $leftoverEquityAccount, $partnerRemainder, $toDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }
      }
    }
  }
  elsif ($businessType == 3)
  {
    my @percentages = @{$args{percentages}};

    my $shareholderTotal = $retainedEarnings;
    for (my $count=0; $count < scalar @eqAccounts; $count++)
    {
      my $percentage = $percentages[$count];
      # divide percentage by 100 to get the actual decimal amount; keep 6 decimal places
      $percentage /= 100.0;
      $percentage = sprintf("%.6f", $percentage);

      my $shareAmount = "0.00";
      if ($retainedEarnings ne "0.00")
      {
        $shareAmount = $retainedEarnings * $percentage;
        if ($shareAmount < 0)
        {
          $shareAmount += "0.005";
        }
        else
        {
          $shareAmount -= "0.005";
        }
        $shareAmount = sprintf("%.2f", $shareAmount);
      }
      else
      {
        $shareAmount = "0.00";
      }

      $shareholderTotal -= $shareAmount;
      $shareholderTotal = sprintf("%.2f", $shareholderTotal);

      # obtain the equity account balance, to add to the total it already has
      $sth = $self->{dbHandle}->read(sql => "SELECT balance FROM acct_bal_tb WHERE acct_num = ? AND sub_acct_num = '-2' AND bal_date = ? AND ledger_code = ?", plug => [ $eqAccounts[$count], $toDate, $ledgerCode ]);
      if ($self->{dbHandle}->didErrorOccur)
      {
        $self->setError(errorString => sprintf($self->{langObj}->map("errorLookingUpBalanceForShareholderEquityAccount"), $eqAccounts[$count]) . "<br>\n" . $self->{dbHandle}->errorMessage);
        return 0;
      }
      my $shareholderBalanceAmount = "0.00";
      if (defined (@info = $sth->fetchrow_array) && defined $info[0])
      {
        # add to existing balance entries
        $shareholderBalanceAmount = $info[0];
        $shareholderBalanceAmount += $shareAmount;
        $shareholderBalanceAmount = sprintf("%.2f", $shareholderBalanceAmount);

        $sth = $self->{dbHandle}->write(sql => "UPDATE acct_bal_tb SET balance = ? WHERE acct_num = ? AND sub_acct_num = '-2' AND bal_date = ? AND ledger_code = ?", plug => [ $shareholderBalanceAmount, $eqAccounts[$count], $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorUpdatingRetainedEarningsBalanceForShareholderEquityAccount"), $eqAccounts[$count], $shareholderBalanceAmount, $toDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }

        $sth = $self->{dbHandle}->read(sql => "SELECT balance FROM acct_bal_tb WHERE acct_num = ? AND sub_acct_num = '-1' AND bal_date = ? AND ledger_code = ?", plug => [ $eqAccounts[$count], $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorLookingUpBalanceForShareholderEquitySubAccount"), $eqAccounts[$count], $businessType) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }
        if (defined (@info = $sth->fetchrow_array) && defined $info[0])
        {
          $shareholderBalanceAmount = $info[0];
          $shareholderBalanceAmount += $shareAmount;
          $shareholderBalanceAmount = sprintf("%.2f", $shareholderBalanceAmount);
        }
        else
        {
          $self->setError(errorString => $self->{langObj}->map("errorNoBalancedReturnedForShareholderEquitySubAccount"));
          return 0;
        }

        $sth = $self->{dbHandle}->write(sql => "UPDATE acct_bal_tb SET balance = ? WHERE acct_num = ? AND sub_acct_num = '-1' AND bal_date = ? AND ledger_code = ?", plug => [ $shareholderBalanceAmount, $eqAccounts[$count], $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorUpdatingRetainedEarningsBalanceForShareholderEquitySubAccount"), $eqAccounts[$count], $shareholderBalanceAmount, $toDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }
      }
      else
      {
        # create new balance entries
        $sth = $self->{dbHandle}->write(sql => "INSERT INTO acct_bal_tb (acct_num, sub_acct_num, balance, bal_date, ledger_code) VALUES (?, '-2', ?, ?, ?)", plug => [ $eqAccounts[$count], $shareAmount, $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorCreatingRetainedEarningsBalanceForShareholderEquityAccount"), $eqAccounts[$count], $shareAmount, $toDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }
        $sth = $self->{dbHandle}->write(sql => "INSERT INTO acct_bal_tb (acct_num, sub_acct_num, balance, bal_date, ledger_code) VALUES (?, '-1', ?, ?, ?)", plug => [ $eqAccounts[$count], $shareAmount, $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorCreatingRetainedEarningsBalanceForShareholderEquitySubAccount"), $eqAccounts[$count], $shareAmount, $toDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }
      }
    }

    if ($shareholderTotal ne "0.00")
    {
      # obtain the equity account balance, to add to the total it already has
      $sth = $self->{dbHandle}->read(sql => "SELECT balance FROM acct_bal_tb WHERE acct_num = ? AND sub_acct_num = '-2' AND bal_date = ? AND ledger_code = ?", plug => [ $leftoverEquityAccount, $toDate, $ledgerCode ]);
      if ($self->{dbHandle}->didErrorOccur)
      {
        $self->setError(errorString => sprintf($self->{langObj}->map("errorLookingUpBalanceForShareholderLeftoverEquityAccount"), $leftoverEquityAccount) . "<br>\n" . $self->{dbHandle}->errorMessage);
        return 0;
      }

      my $leftoverShareholderAmount = "0.00";
      if (defined (@info = $sth->fetchrow_array) && defined $info[0])
      {
        # add to existing balance entries
        $leftoverShareholderAmount = $info[0];
        $leftoverShareholderAmount += $shareholderTotal;
        $leftoverShareholderAmount = sprintf("%.2f", $leftoverShareholderAmount);

        $sth = $self->{dbHandle}->write(sql => "UPDATE acct_bal_tb SET balance = ? WHERE acct_num = ? AND sub_acct_num = '-2' AND bal_date = ? AND ledger_code = ?", plug => [ $leftoverShareholderAmount, $leftoverEquityAccount, $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorUpdatingRetainedEarningsBalanceForLeftoverEquityAccount"), $leftoverEquityAccount, $leftoverShareholderAmount, $toDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }

        $sth = $self->{dbHandle}->read(sql => "SELECT balance FROM acct_bal_tb WHERE acct_num = ? AND sub_acct_num = '-1' AND bal_date = ? AND ledger_code = ?", plug => [ $leftoverEquityAccount, $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorLookingUpBalanceForLeftoverShareholderEquitySubAccount"), $leftoverEquityAccount, $businessType) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }
        if (defined (@info = $sth->fetchrow_array) && defined $info[0])
        {
          $leftoverShareholderAmount = $info[0];
          $leftoverShareholderAmount += $shareholderTotal;
          $leftoverShareholderAmount = sprintf("%.2f", $leftoverShareholderAmount);
        }
        else
        {
          $self->setError(errorString => $self->{langObj}->map("errorNoBalancedReturnedForLeftoverEquitySubAccount"));
          return 0;
        }

        $sth = $self->{dbHandle}->write(sql => "UPDATE acct_bal_tb SET balance = ? WHERE acct_num = ? AND sub_acct_num = '-1' AND bal_date = ? AND ledger_code = ?", plug => [ $leftoverShareholderAmount, $leftoverEquityAccount, $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorUpdatingRetainedEarningsBalanceForLeftoverEquitySubAccount"), $leftoverEquityAccount, $leftoverShareholderAmount, $toDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }
      }
      else
      {
        # create new balance entries
        $sth = $self->{dbHandle}->write(sql => "INSERT INTO acct_bal_tb (acct_num, sub_acct_num, balance, bal_date, ledger_code) VALUES (?, '-2', ?, ?, ?)", plug => [ $leftoverEquityAccount, $shareholderTotal, $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorCreatingRetainedEarningsBalanceForLeftoverEquityAccount"), $leftoverEquityAccount, $shareholderTotal, $toDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }
        $sth = $self->{dbHandle}->write(sql => "INSERT INTO acct_bal_tb (acct_num, sub_acct_num, balance, bal_date, ledger_code) VALUES (?, '-1', ?, ?, ?)", plug => [ $leftoverEquityAccount, $shareholderTotal, $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorCreatingRetainedEarningsBalanceForLeftoverEquitySubAccount"), $leftoverEquityAccount, $shareholderTotal, $toDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }
      }
    }
  }
  else
  {
    $sth = $self->{dbHandle}->read(sql => "SELECT balance FROM acct_bal_tb WHERE acct_num = ? AND sub_acct_num = '-2' AND bal_date = ? AND ledger_code = ?", plug => [ $eqAccounts[0], $toDate, $ledgerCode ]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError(errorString => sprintf($self->{langObj}->map("errorLookingUpBalanceForEquityAccount"), $eqAccounts[0]) . "<br>\n" . $self->{dbHandle}->errorMessage);
      return 0;
    }

    my $balanceAmount = "0.00";
    if (defined (@info = $sth->fetchrow_array) && defined $info[0])
    {
      $balanceAmount = $info[0];
      # add to the existing balance amount using Retained Earnings
      $balanceAmount += $retainedEarnings;
      $balanceAmount = sprintf("%.2f", $balanceAmount);

      $sth = $self->{dbHandle}->write(sql => "UPDATE acct_bal_tb SET balance = ? WHERE acct_num = ? AND sub_acct_num = '-2' AND bal_date = ? AND ledger_code = ?", plug => [ $balanceAmount, $eqAccounts[0], $toDate, $ledgerCode ]);
      if ($self->{dbHandle}->didErrorOccur)
      {
        $self->setError(errorString => sprintf($self->{langObj}->map("errorUpdatingRetainedEarningsBalanceForEquityAccount"), $eqAccounts[0], $balanceAmount, $toDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
        return 0;
      }

      $sth = $self->{dbHandle}->read(sql => "SELECT balance FROM acct_bal_tb WHERE acct_num = ? AND sub_acct_num = '-1' AND bal_date = ? AND ledger_code = ?", plug => [ $eqAccounts[0], $toDate, $ledgerCode ]);
      if ($self->{dbHandle}->didErrorOccur)
      {
        $self->setError(errorString => sprintf($self->{langObj}->map("errorLookingUpBalanceForEquitySubAccount"), $eqAccounts[0]) . "<br>\n" . $self->{dbHandle}->errorMessage);
        return 0;
      }
      my @info2 = $sth->fetchrow_array;
      if (defined @info2 && defined $info2[0])
      {
        # update the sub-account -1 balance
        my $subBalanceAmount = $info2[0];
        $subBalanceAmount += $retainedEarnings;
        $subBalanceAmount = sprintf("%.2f", $subBalanceAmount);

        $sth = $self->{dbHandle}->write(sql => "UPDATE acct_bal_tb SET balance = ? WHERE acct_num = ? AND sub_acct_num = '-1' AND bal_date = ? AND ledger_code = ?", plug => [ $subBalanceAmount, $eqAccounts[0], $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorUpdatingRetainedEarningsBalanceForEquitySubAccount"), $eqAccounts[0], $subBalanceAmount, $toDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }
      }
      else
      {
        # create the sub-account -1 balance
        $sth = $self->{dbHandle}->write(sql => "INSERT INTO acct_bal_tb (acct_num, sub_acct_num, balance, bal_date, ledger_code) VALUES (?, '-1', ?, ?, ?)", plug => [ $eqAccounts[0], $retainedEarnings, $toDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorCreatingRetainedEarningsBalanceForEquitySubAccount"), $eqAccounts[0], $retainedEarnings, $toDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }
      }
    }
    else
    {
      # create the necessary balance amount using Retained Earnings
      my $balanceSQL = "INSERT INTO acct_bal_tb (acct_num, sub_acct_num, balance, bal_date, ledger_code) VALUES (?, '-2', ?, ?, ?); ";
      $balanceSQL .= "INSERT INITO acct_bal_tb (acct_num, sub_acct_num, balance, bal_date, ledger_code) VALUES (?, '-1', ?, ?, ?)";
      $sth = $self->{dbHandle}->write(sql => $balanceSQL, plug => [ $eqAccounts[0], $retainedEarnings, $toDate, $ledgerCode, $eqAccounts[0], $retainedEarnings, $toDate, $ledgerCode ]);
      if ($self->{dbHandle}->didErrorOccur)
      {
        $self->setError(errorString => sprintf($self->{langObj}->map("errorCreatingRetainedEarningsBalanceForEquityAccount"), $eqAccounts[0], $retainedEarnings, $toDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
        return 0;
      }
    }
  }

  # update the next fiscal year or closeout period
  if ($checkDateRange ne "0")
  {
    if ($workingYearEnd eq $toDate)
    {
      my $nextFiscalYear = $currentFiscalYear + 1;

      $sth = $self->{dbHandle}->read(sql => "SELECT year, closed FROM known_years_tb WHERE year = ? AND ledger_code = ?", plug => [ $nextFiscalYear, $ledgerCode ]);
      if ($self->{dbHandle}->didErrorOccur)
      {
        $self->setError(errorString => sprintf($self->{langObj}->map("errorLookingUpNextFiscalYear"), $nextFiscalYear) . "<br>\n" . $self->{dbHandle}->errorMessage);
        return 0;
      }
      if (defined (@info = $sth->fetchrow_array) && defined $info[0])
      {
        # update the next current fiscal year entry to set it active
        $sth = $self->{dbHandle}->write(sql => "UPDATE known_years_tb SET closed = false WHERE year = ? AND ledger_code = ?", plug => [ $nextFiscalYear, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorUpdatingNextFiscalYearRecord"), $nextFiscalYear) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }
      }
      else
      {
        # create a new fiscal year entry, if needed
        my $result;
        my $errorString;
        my $beginDate;
        my $endDate;

        $beginDate = $self->{sessionObj}->{store}->{ledgers}->{$ledgerName}->start(month => 1, fiscalYear => $nextFiscalYear);
        $endDate = $self->{sessionObj}->{store}->{ledgers}->{$ledgerName}->end(month => 12, fiscalYear => $nextFiscalYear);    

        $sth = $self->{dbHandle}->write(sql => "INSERT INTO known_years_tb (year, closed, start_date, end_date, ledger_code) VALUES (?, false, ?, ?, ?)", plug => [ $nextFiscalYear, $beginDate, $endDate, $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur)
        {
          $self->setError(errorString => sprintf($self->{langObj}->map("errorCreatingNextFiscalYear"), $nextFiscalYear, $beginDate, $endDate) . "<br>\n" . $self->{dbHandle}->errorMessage);
          return 0;
        }

        # check to see if we are working with a calendar year
        if ($beginDate =~ /01.01.\d{4}/ && $endDate =~ /12.31.\d{4}/)
        {
          # calendar year
          my $quarterStart;
          my $quarterEnd;
          for (my $quarter=1; $quarter <= 4; $quarter++)
          {
            if ($quarter == 1)
            {
              $quarterStart = "01/01/$nextFiscalYear";
              $quarterEnd = "03/31/$nextFiscalYear";
            }
            elsif ($quarter == 2)
            {
              $quarterStart = "04/01/$nextFiscalYear";
              $quarterEnd = "06/30/$nextFiscalYear";
            }
            elsif ($quarter == 3)
            {
              $quarterStart = "07/01/$nextFiscalYear";
              $quarterEnd = "09/30/$nextFiscalYear";
            }
            elsif ($quarter == 4)
            {
              $quarterStart = "10/01/$nextFiscalYear";
              $quarterEnd = "12/31/$nextFiscalYear";
            }

            $sth = $self->{dbHandle}->write(sql => "UPDATE known_years_tb SET q" . $quarter . "_start = ?, q" . $quarter . "_end = ? WHERE year = ? AND ledger_code = ?", plug => [ $quarterStart, $quarterEnd, $nextFiscalYear, $ledgerCode ]);
            if ($self->{dbHandle}->didErrorOccur)
            {
              $self->setError(errorString => sprintf($self->{langObj}->map("errorUpdatingNextFiscalYearQuarter"), $nextFiscalYear, $quarter, $quarterStart, $quarterEnd) . "<br>\n" . $self->{dbHandle}->errorMessage);
              return 0;
            }
          }
        }
        else
        {
          # non-calendar year
          my $quarterStart;
          my $quarterEnd;
          for (my $quarter=1; $quarter <= 4; $quarter++)
          {
            $quarterStart = $beginDate;
            if ($quarter != 4)
            {
              ($result, $quarterEnd, $errorString) = $self->{methods}->getDateOffset(date => $beginDate, offset => "-1 day 3 months");
              if ($result == 0 || $errorString ne "")
              {
                $self->setError(errorString => sprintf($self->{langObj}->map("errorObtainingQuarterEndOffset"), $quarter) . "<br>\n" . $errorString);
                return 0;
              }

              ($result, $beginDate, $errorString) = $self->{methods}->getDateOffset(date => $quarterEnd, offset => "1 day");
              if ($result == 0 || $errorString ne "")
              {
                $self->setError(errorString => sprintf($self->{langObj}->map("errorObtainingNextBeginDateOffset"), $quarter) . "<br>\n" . $errorString);
                return 0;
              }
            }
            else
            {
              $quarterEnd = $endDate;
            }

            $sth = $self->{dbHandle}->write(sql => "UPDATE known_years_tb SET q" . $quarter . "_start = ?, q" . $quarter . "_end = ? WHERE year = ? AND ledger_code = ?", plug => [ $quarterStart, $quarterEnd, $nextFiscalYear, $ledgerCode ]);
            if ($self->{dbHandle}->didErrorOccur)
            {
              $self->setError(errorString => sprintf($self->{langObj}->map("errorUpdatingNextFiscalYearQuarter"), $nextFiscalYear, $quarter, $quarterStart, $quarterEnd) . "<br>\n" . $self->{dbHandle}->errorMessage);
              return 0;
            }
          }
        }
      }

      # mark the old current fiscal year closed
      $sth = $self->{dbHandle}->write(sql => "UPDATE known_years_tb SET closed = true WHERE year = ? AND ledger_code = ?", plug => [ $currentFiscalYear, $ledgerCode ]);
      if ($self->{dbHandle}->didErrorOccur)
      {
        $self->setError(errorString => sprintf($self->{langObj}->map("errorUpdatingClosedFiscalYearRecord"), $currentFiscalYear) . "<br>\n" . $self->{dbHandle}->errorMessage);
        return 0;
      }

      # update the variables_tb table to reflect the next fiscal year
      $sth = $self->{dbHandle}->write(sql => "UPDATE variables_tb SET working_year = ? WHERE ledger_code = ?", plug => [ $nextFiscalYear, $ledgerCode ]);
      if ($self->{dbHandle}->didErrorOccur)
      {
        $self->setError(errorString => sprintf($self->{langObj}->map("errorUpdatingNewWorkingYear"), $nextFiscalYear, $ledgerCode) . "<br>\n" . $self->{dbHandle}->errorMessage);
        return 0;
      }
    }
  }

  # commit the work
  if ($args{commit} == 1)
  {
    $sth = $self->{dbHandle}->commit;
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError(errorString => $self->{langObj}->map("errorCommittingTransaction") . "<br>\n" . $self->{dbHandle}->errorMessage);
      return 0;
    }
  }

  # we successfully closed the books
  return 1;
}

=item int reopenLedger(ledgerCode, date, equityAccounts)

 Reopens a ledger on the given year or date period.
 Returns 1 on success, 0 on errors and sets the errorMessage.

=cut

sub reopenLedger
{
  my $self = shift;
  my %args = ( ledgerCode => "", reopenDate => "", commit => 1, @_ );
  my $ledgerCode = $args{ledgerCode};
  my $reopenDate = $args{reopenDate};
  my $lookupYear;
  my $toDate;

  # validate input
  if (not defined $ledgerCode || $ledgerCode eq "")
  {
    $self->setError(errorString => sprintf($self->{langObj}->map("invalidArgument"), "ledgerCode", $ledgerCode));
    return 0;
  }
  if (not defined $reopenDate || $reopenDate eq "")
  {
    $self->setError(errorString => sprintf($self->{langObj}->map("invalidArgument"), "reopenDate", $reopenDate));
    return 0;
  }

  if ($reopenDate =~ /(-|\/)/)
  {
    (my $result, my $fixedReopenDate, my $errorString) = $self->{methods}->fixupDate(date => $reopenDate, seperator => '-');
    if ($result == 0)
    {
      $self->setError(errorString => sprintf($self->{langObj}->map("errorFixingUpDate"), $reopenDate) . "\n" . $errorString);
      return 0;
    }
    $reopenDate = $fixedReopenDate;  # assign the good date to the original variable
    $toDate = $reopenDate;

    ($lookupYear, my $tempMonth, my $tempDay) = split(/-/g, $reopenDate);
  }
  elsif ($reopenDate =~ /^\d+$/)
  {
    $lookupYear = $reopenDate;

    # get the fiscal year end date for the lookupYear
    my $sth = $self->{dbHandle}->read(sql => "SELECT end_date FROM known_years_tb WHERE year = ? AND ledger_code = ?", plug => [ $lookupYear, $ledgerCode ]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError(errorString => sprintf($self->{langObj}->map("errorLookingUpFiscalYearEnd"), $lookupYear) . "<br>\n" . $self->{dbHandle}->errorMessage);
      return 0;
    }
    my @info = $sth->fetchrow_array;
    if (defined @info && defined $info[0])
    {
      $toDate = $info[0];
    }
    else
    {
      $self->setError(errorString => sprintf($self->{langObj}->map("errorLookingUpFiscalYearEnd"), $lookupYear));
      return 0;
    }
  }
  else
  {
    $self->setError(errorString => sprintf($self->{langObj}->map("invalidArgument"), "reopenDate", $reopenDate));
    return 0;
  }

  # remove any balance table entries that exist on or after the given reopenDate period
  my $sth = $self->{dbHandle}->write(sql => "DELETE FROM acct_bal_tb WHERE bal_date >= ? AND ledger_code = ?", plug => [ $toDate, $ledgerCode ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => sprintf($self->{langObj}->map("errorRemovingAccountBalances"), $lookupYear) . "<br>\n" . $self->{dbHandle}->errorMessage);
    return 0;
  }

  my $sth = $self->{dbHandle}->write(sql => "UPDATE known_years_tb SET closed = false WHERE year >= ? AND ledger_code = ?", plug => [ $lookupYear, $ledgerCode ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => sprintf($self->{langObj}->map("errorOpeningFiscalYears"), $lookupYear) . "<br>\n" . $self->{dbHandle}->errorMessage);
    return 0;
  }

  # determine whether the lookupYear is less than the least known_years entry for the ledgerCode
  # if it is, do not update the working_year to lookupYear, but set it to the least known_years entry
  my $sth = $self->{dbHandle}->read(sql => "SELECT MIN(year) FROM known_years_tb WHERE ledger_code = ?", plug => [ $ledgerCode ]);
  my @info = $sth->fetchrow_array;
  if ((not defined @info) || ((defined @info) && (not defined $info[0])))
  {
    $self->setError(errorString => sprintf($self->{langObj}->map("errorLookingUpMinimumKnownYear"), $ledgerCode));
    return 0;
  }
  my $minimumKnownYear = $info[0];
  if ($minimumKnownYear > $lookupYear)
  {
    my $sth = $self->{dbHandle}->write(sql => "UPDATE variables_tb SET working_year = ? WHERE ledger_code = ?", plug => [ $minimumKnownYear, $ledgerCode ]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError(errorString => $self->{langObj}->map("errorUpdatingWorkingYear") . "<br>\n" . $self->{dbHandle}->errorMessage);
      return 0;
    }
  }
  else
  {
    my $sth = $self->{dbHandle}->write(sql => "UPDATE variables_tb SET working_year = ? WHERE ledger_code = ?", plug => [ $lookupYear, $ledgerCode ]);
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError(errorString => $self->{langObj}->map("errorUpdatingWorkingYear") . "<br>\n" . $self->{dbHandle}->errorMessage);
      return 0;
    }
  }

  # commit the work
  if ($args{commit} == 1)
  {
    $sth = $self->{dbHandle}->commit;
    if ($self->{dbHandle}->didErrorOccur)
    {
      $self->setError(errorString => $self->{langObj}->map("errorCommittingTransaction") . "<br>\n" . $self->{dbHandle}->errorMessage);
      return 0;
    }
  }

  # we successfully re-opened the books
  return 1;
}


=item Objects::Ledger[] getLedgers(ledgerCode =>, ledgerName =>)

    Returns an array of Ledger Objects matching the parameters passed.
    No params means select all ledgers.  Empty array is returned if
    no ledgers exist or match criteria.
    NOTE: Will not error on an invalid ledger.  You must do an
    ->isValid on each returned ledgerObj to decide what to do with it.
    Undef returned on errors.

=cut
sub getLedgers
{
  my $self = shift;
  my %args = ( ledgerCode => "", ledgerName => "", @_ ) ;
  my @Ledgers = ();
  my $where = "";

  if ($args{ledgerCode} ne "" || $args{ledgerName} ne "")
  {
    $where .= " AND ";
    if ($args{ledgerCode} ne "" && $args{ledgerName} ne "")
    {
      $where .= " ( ledger_tb.name = '" . $args{ledgerName} . "'";
      $where .= " OR ledger_tb.code = '" . $args{ledgerCode} . "' ) ";
    }
    elsif ($args{ledgerName} ne "")
    {
      $where .= " ledger_tb.name = '" . $args{ledgerName} . "'";
    }
    else
    {
      $where .= " ledger_tb.code = '" . $args{ledgerCode} . "'";
    }
  }

  my $sql = "SELECT ledger_tb.name, ledger_tb.code, ledger_tb.color";
  $sql .= ", ledger_tb.cash_or_accrual ";
  $sql .= ", business_type_tb.name, ledger_tb.business_type ";
  $sql .= ", offset1 ";
  $sql .= ", start1 ";
  $sql .= ", start2 ";
  $sql .= ", start3 ";
  $sql .= ", start4 ";
  $sql .= ", start5 ";
  $sql .= ", start6 ";
  $sql .= ", start7 ";
  $sql .= ", start8 ";
  $sql .= ", start9 ";
  $sql .= ", start10 ";
  $sql .= ", start11 ";
  $sql .= ", start12 ";
  $sql .= "FROM ledger_tb, business_type_tb ";
  $sql .= "WHERE ledger_tb.business_type = business_type_tb.num ".$where;
  $sql .= "ORDER BY ledger_tb.name, ledger_tb.code";
  my $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError($sql."<br>\n". $self->{dbHandle}->errorMessage);
    return undef;
  }
  while (my @info = $sth->fetchrow_array)
  {
    my %knownYears = $self->getKnownYears(ledgerCode => $info[1]);
    if ($self->didErrorOccur)
    {
      $self->setError(" Sent ledger '$args{ledgerCode}' to function getKnownYears<br>\n");
      return undef;
    }
    my $cashOrAccrual = "";
    if ($info[3] =~ /^t|true|1$/)  #!!!!find DB independent way to do this!
    {  $cashOrAccrual = "cash";}
    elsif ($info[3] =~ /^f|false|0$/)
    { $cashOrAccrual = "accrual";}
    else
    { $cashOrAccrual = "error!";}

    my @start = ( $info[6], $info[7], $info[8], $info[9], $info[10], $info[11], $info[12], $info[13], $info[14], $info[15], $info[16], $info[17], $info[18], $info[19]);

    my $ledgerObj = Portal::Accounting::Objects::Ledger->new( langObj => $self->{langObj},
                    name => $info[0], code => $info[1], color => $info[2],
                    cashOrAccrual => $cashOrAccrual,
                    businessType => $info[5],
                    businessTypeName => $info[4],
                    knownYears => \%knownYears,
                    start => \@start,
                    );

    $Ledgers[++$#Ledgers] = $ledgerObj;
  }

  return @Ledgers;
}

=item Objects::Ledger getLedger(ledgerCode =>)

    Returns the Ledger Object matching the specified ledgerCode or
    undef if not found.

    You need to verify the validity of the returned object yourself.

=cut
sub getLedger
{
  my $self = shift;
  my %args = ( ledgerCode => "", @_ ) ;
  my $ledger = undef;
  my $where = "";

  if ($args{ledgerCode} ne "")
  {
    $where .= " AND ";
    $where .= " ledger_tb.code = '" . $args{ledgerCode} . "'";
  }
  else
  {
    $self->missing("ledgerCode");
    $self->postfixError($self->genErrorString("all"));
    return undef;
  }

  my $sql = "SELECT ledger_tb.name, ledger_tb.code, ledger_tb.color";
  $sql .= ", ledger_tb.cash_or_accrual ";
  $sql .= ", business_type_tb.name, ledger_tb.business_type ";
  $sql .= ", offset1 ";
  $sql .= ", start1 ";
  $sql .= ", start2 ";
  $sql .= ", start3 ";
  $sql .= ", start4 ";
  $sql .= ", start5 ";
  $sql .= ", start6 ";
  $sql .= ", start7 ";
  $sql .= ", start8 ";
  $sql .= ", start9 ";
  $sql .= ", start10 ";
  $sql .= ", start11 ";
  $sql .= ", start12 ";
  $sql .= "FROM ledger_tb, business_type_tb ";
  $sql .= "WHERE ledger_tb.business_type = business_type_tb.num ".$where;
  $sql .= "ORDER BY ledger_tb.name, ledger_tb.code";
  my $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError($sql."<br>\n". $self->{dbHandle}->errorMessage);
    return undef;
  }
  my @info = $sth->fetchrow_array;
  if (defined $info[0])
  {
    my %knownYears = $self->getKnownYears(ledgerCode => $info[1]);
    if ($self->didErrorOccur)
    {
      $self->setError(" Sent ledger '$args{ledgerCode}' to function getKnownYears<br>\n");
      return undef;
    }
    my $cashOrAccrual = "";
    if ($info[3] =~ /^t|true|1$/)  #!!!!find DB independent way to do this!
    {  $cashOrAccrual = "cash";}
    elsif ($info[3] =~ /^f|false|0$/)
    { $cashOrAccrual = "accrual";}
    else
    { $cashOrAccrual = "error!";}

    my @start = ( $info[6], $info[7], $info[8], $info[9], $info[10], $info[11], $info[12], $info[13], $info[14], $info[15], $info[16], $info[17], $info[18], $info[19]);

    $ledger = Portal::Accounting::Objects::Ledger->new( langObj => $self->{langObj},
                    name => $info[0], code => $info[1], color => $info[2],
                    cashOrAccrual => $cashOrAccrual,
                    businessType => $info[5],
                    businessTypeName => $info[4],
                    knownYears => \%knownYears,
                    start => \@start,
                    );
  }
  else
  {
    return undef;
  }

  return $ledger;
}

=item Objects::Ledger getValidLedger()
  
  Since ledgers need to show up whether they have all valid data getLedger will return all of them.
  This routine will check the ledger returned and error if it is not valid.
  
=cut
sub getValidLedger
{
  my $self = shift;
  my $ledger = $self->getLedger(@_);
  if (not defined $ledger)
  { return undef; }
  if ($ledger->isValid)
  {  return $ledger; }
  return undef;
}

=head2 Hash getBusinessTypeNames()

    This routine selects the business type names from the database and returns them as an hash
    of number = name values where the name is looked up from the Language Object so as to be
    Internationalized.

=cut
sub getBusinessTypeNames
{
  my $self = shift;
  my $errStr = "Portal::Accounting::DB::Ledger->getBusinessTypes() - Error!<br>\n";

  my $index = 0;
  my @info;
  my $sth;
  my %types;

  $sth = $self->{dbHandle}->read(sql => "SELECT name, num FROM business_type_tb ORDER BY num");
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->{error} = 1;
    $self->{errorString} = $errStr . $self->{dbHandle}->errorMessage;
    return undef;
  }
  while (@info = $sth->fetchrow_array)
  {
    $types{$info[1]} = $self->{langObj}->map($info[0]);
  }

  return %types;
}

=head2 bool isValidData(field => businessType, data => "Corporation", ledger => $ledger)

=cut
sub isValidData
{
  my $self = shift;
  my $errStr = "Portal::Accounting::DB::Ledger->isValidData() - Error!<br>\n";

  return undef;
}

=head2 SubEntityObject getSubEntities(ledger =>)

=cut
sub getSubEntities
{
  my $self = shift;
  my $errStr = "Portal::Accounting::DB::Ledger->getSubEntities() - Error!<br>\n";

  return undef;
}

=head2 bool addSubEntity(ledger=>, code =>, description =>)

=cut
sub addSubEntity
{
  my $self = shift;
  my $errStr = "Portal::Accounting::DB::Ledger->addSubEntity() - Error!<br>\n";

  return undef;
}

=head2 bool deleteSubEntity(ledger=>, code =>, description =>)

=cut
sub deleteSubEntity
{
  my $self = shift;
  my $errStr = "Portal::Accounting::DB::Ledger->deleteSubEntity() - Error!<br>\n";

  return undef;
}

=item int doesSystemBalance()

     requires: ledgerCode, beginDate, endDate
     optional: if beginDate/endDate are not given, the balances will be calculated on the remaining date value.
               If neither date value is given, the balance will be calculated on the entire set of transactions
               for the given ledgerCode.
     returns: 1 if system balances, -1 if debits are greater than credits, -2 if credits are greater than debits,
              or 0 if an error occurred
    summary: This method sums all the debits and credits for the transactions in the debitcredit_tb table for the
             given ledgerCode and dates and compares them to see if they are the same.  It will return different
             error codes based on whether debits are greater than credits or vice versa.
=cut
sub doesSystemBalance
{
  my $self = shift;
  my %args = ( ledgerCode => "", beginDate => "", endDate => "", @_ );
  my $ledgerCode = $args{ledgerCode};
  my $beginDate = $args{beginDate};
  my $endDate = $args{endDate};
  my $sth;
  my @info;
  my $sqlString = "SELECT SUM(amount) FROM debitcredit_tb WHERE ledger_code = ? AND ";
  my $sqlWhere = "";
  my @plug = ();
  my $sumDebits = "0.00";
  my $sumCredits = "0.00";
  my $result;

  if (not defined $ledgerCode || length $ledgerCode == 0)
  {
    $self->setError(errorString => sprintf($self->{langObj}->map("invalidArgument"), "ledgerCode", $ledgerCode));
    return 0;
  }

  # put the valid ledgerCode onto the plug array
  push @plug, $ledgerCode;

  # debits
  if ((defined $beginDate) && ($beginDate =~ /(\d{4}-?\d{1,2}-?\d{1,2})|(\d{1,2}-?\d{1,2}-?\d{4})/) && (defined $endDate) && ($endDate =~ /(\d{4}-?\d{1,2}-?\d{1,2})|(\d{1,2}-?\d{1,2}-?\d{4})/))
  {
    $sqlWhere = "(trans_date BETWEEN ? AND ?) AND dorc = true";
    push @plug, $beginDate, $endDate;
  }
  elsif (((not defined $beginDate) || (length $beginDate == 0)) && ((defined $endDate) && ($endDate =~ /(\d{4}-?\d{1,2}-?\d{1,2})|(\d{1,2}-?\d{1,2}-?\d{4})/)))
  {
    $sqlWhere = "trans_date <= ? AND dorc = true";
    push @plug, $endDate;
  }
  elsif (((not defined $endDate) || (length $endDate == 0)) && ((defined $beginDate) && ($beginDate =~ /(\d{4}-?\d{1,2}-?\d{1,2})|(\d{1,2}-?\d{1,2}-?\d{4})/)))
  {
    $sqlWhere = "trans_date >= ? AND dorc = true";
    push @plug, $beginDate;
  }
  elsif (((not defined $beginDate) || (length $beginDate == 0)) && ((not defined $endDate) || (length $endDate == 0)))
  {
    $sqlWhere = "dorc = true";
  }
  else
  {
    $self->setError(errorString => sprintf($self->{langObj}->map("invalidDebitDate"), $beginDate, $endDate));
    return 0;
  }
  # retrieve the sum of all debit amounts
  $sth = $self->{dbHandle}->read(sql => $sqlString . $sqlWhere, plug => \@plug);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => $self->{langObj}->map("errorSummingDebits") . "<br>\n" . $self->{dbHandle}->errorMessage);
    return 0;
  }
  @info = $sth->fetchrow_array;
  if (defined @info && defined $info[0])
  {
    $sumDebits = $info[0];
  }
  else
  {
    $sumDebits = "0.00";
  }

  # refresh the plug array for the credits
  @plug = ();
  push @plug, $ledgerCode;

  # credits
  if ((defined $beginDate) && ($beginDate =~ /(\d{4}-?\d{1,2}-?\d{1,2})|(\d{1,2}-?\d{1,2}-?\d{4})/) && (defined $endDate) && ($endDate =~ /(\d{4}-?\d{1,2}-?\d{1,2})|(\d{1,2}-?\d{1,2}-?\d{4})/))
  {
    $sqlWhere = "(trans_date BETWEEN ? AND ?) AND dorc = false";
    push @plug, $beginDate, $endDate;
  }
  elsif (((not defined $beginDate) || (length $beginDate == 0)) && ((defined $endDate) && ($endDate =~ /(\d{4}-?\d{1,2}-?\d{1,2})|(\d{1,2}-?\d{1,2}-?\d{4})/)))
  {
    $sqlWhere = "trans_date <= ? AND dorc = false";
    push @plug, $endDate;
  }
  elsif (((not defined $endDate) || (length $endDate == 0)) && ((defined $beginDate) && ($beginDate =~ /(\d{4}-?\d{1,2}-?\d{1,2})|(\d{1,2}-?\d{1,2}-?\d{4})/)))
  {
    $sqlWhere = "trans_date >= ? AND dorc = false";
    push @plug, $beginDate;
  }
  elsif (((not defined $beginDate) || (length $beginDate == 0)) && ((not defined $endDate) || (length $endDate == 0)))
  {
    $sqlWhere = "dorc = false";
  }
  else
  {
    $self->setError(errorString => sprintf($self->{langObj}->map("invalidCreditDate"), $beginDate, $endDate));
    return 0;
  }
  # retrieve the sum of all credit amounts
  $sth = $self->{dbHandle}->read(sql => $sqlString . $sqlWhere, plug => \@plug);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => $self->{langObj}->map("errorSummingCredits") . "<br>\n" . $self->{dbHandle}->errorMessage);
    return 0;
  }
  @info = $sth->fetchrow_array;
  if (defined @info && defined $info[0])
  {
    $sumCredits = $info[0];
  }
  else
  {
    $sumCredits = "0.00";
  }

  # make sure both values are formatted correctly
  $sumDebits = sprintf("%.2f", $sumDebits);
  $sumCredits = sprintf("%.2f", $sumCredits);

  # compare to make sure that debits == credits
  if ($sumDebits eq $sumCredits)
  {
    # the system balances
    $result = 1;
  }
  elsif ($sumDebits > $sumCredits)
  {
    $result = -1;
  }
  elsif ($sumDebits < $sumCredits)
  {
    $result = -2;
  }

  return $result;
}

=item integer getCurrentFiscalYear(ledgerCode)

     requires: ledgerCode
     optional:
     returns: integer or undef if error
    summary: This method returns the current Fiscal Year for the specified ledger.
=cut
sub getCurrentFiscalYear
{
  my $self = shift;
  my $ledgerCode = shift;
  my $sth;
  my @info;

  if (length $ledgerCode == 0)
  {
    $self->prefixError( "ledgerCode = '$ledgerCode' is invalid!<br>\n");
    return undef;
  }

  $sth = $self->{dbHandle}->read(sql => "SELECT working_year FROM variables_tb WHERE ledger_code = ?", plug => [ $ledgerCode ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError($self->{dbHandle}->errorMessage);
    return undef;
  }
  if (@info = $sth->fetchrow_array)
  { return $info[0]; }
  $self->prefixError("No data returned!<br>\n");
  return undef;
}

=item date getFiscalEnd(ledgerCode)

     requires: ledgerCode
     optional:
     returns: date or undef if error
    summary: This method returns the last day of the current Fiscal Year for the specified ledger.
=cut
sub getFiscalEnd
{
  my $self = shift;
  my $ledgerCode = shift;
  my $sth;
  my @info;
  my $date = undef;

  if (length $ledgerCode == 0)
  {
    $self->prefixError( "ledgerCode = '$ledgerCode' is invalid!<br>\n");
    return undef;
  }

  $sth = $self->{dbHandle}->read(sql => "SELECT (SELECT working_year + 1 FROM variables_tb WHERE ledger_code = ? LIMIT 1)||'-'||start1 FROM ledger_tb WHERE code = ? LIMIT 1", plug => [ $ledgerCode, $ledgerCode ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->prefixError($self->{dbHandle}->errorMessage);
    return undef;
  }
  @info = $sth->fetchrow_array;
  if (!defined $info[0])
  {
    $self->prefixError("No data returned!<br>\n");
    return undef;
  }
  return $date;
}

# getAllKnownYears
# requires: ledgerCode
# optional:
# returns: array of FiscalYearObjects or undef if error
# summary: This method returns an array of all Fiscal Years for the specified ledger in oldest to newest order.
sub getAllKnownYears
{
  my $self = shift;
  my %args = ( @_ );
  my $ledgerCode = $args{ledgerCode};
  my @knownYears = ();
  my $sth;
  my @info;

  if (length $ledgerCode == 0)
  {
    $self->{error} = 1;
    $self->{errorString} = "ledgerCode = '$ledgerCode' is invalid!<br>\n";
    return @knownYears;
  }

  $sth = $self->{dbHandle}->read(sql => "SELECT year, closed, start_date, end_date, q1_start, q1_end, q2_start, q2_end, q3_start, q3_end, q4_start, q4_end, ledger_code FROM known_years_tb WHERE ledger_code = ? ORDER BY year",
                                    plug => [ $ledgerCode ]);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->{error} = 1;
    $self->{errorString} = $self->{dbHandle}->errorMessage;
    return @knownYears;
  }

  while (@info = $sth->fetchrow_array)
  {
    my $fiscalYearObj = Portal::Accounting::Objects::FiscalYearObject->new(year => $info[0], closed => $self->portalVariables->trueFalse->{$info[1]},
                     startDate => $info[2], endDate => $info[3], q1Start => $info[4], q1End => $info[5], q2Start => $info[6], q2End => $info[7],
                     q3Start => $info[8], q3End => $info[9], q4Start => $info[10], q4End => $info[11], ledgerCode => $info[12], langObj => $self->{langObj});
    if ($fiscalYearObj->didErrorOccur)
    {
      $self->{error} = 1;
      $self->{errorString} = $fiscalYearObj->errorMessage;
      return undef;
    }
    $knownYears[++$#knownYears] = $fiscalYearObj;
  }
  if (scalar @knownYears == 0)
  {
    $self->{error} = 1;
    $self->{errorString} = "No data returned!<br>\n";
    return @knownYears;
  }

  return @knownYears;
}

=head2 %years getKnownYears(ledgerCode =>)

    required: ledgerCode
    Returns hash of years as keys with YYYY/MM/DD as data for each key

=cut
sub getKnownYears
{
  my $self = shift;
  my %args = ( ledgerCode => "", @_ );
  my %knownYears = ();
  my $years = 0;

  if (not defined $args{ledgerCode})
  {
    $self->setError("Error! ledgerCode is not defined!<br>\n");
    return undef;
  }
  if ($args{ledgerCode} eq "")
  {
    $self->setError("Error! ledgerCode = '' is invalid!<br>\n");
    return undef;
  }
  my $sth = $self->{dbHandle}->write(sql => "SET datestyle TO 'ISO'");
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError("Error setting datestyle to ISO!<br>\n" . $self->{dbHandle}->errorMessage);
    return undef;
  }
  my $select = "SELECT year, q4_end from known_years_tb where ledger_code = '$args{ledgerCode}'";
  $sth = $self->{dbHandle}->read(sql => $select);
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError("Error in getKnownYears!<br>\n" . $self->{dbHandle}->errorMessage);
    return undef;
  }
  while (my @info = $sth->fetchrow_array)
  {
    $knownYears{$info[0]} = $info[1];
    $years++;
  }
  if ($years == 0)
  {
    $self->setError( "Error in getKnownYears! No known years returned from database for ledger code '$args{ledgerCode}'.<br>\n");
    return undef;
  }
  return %knownYears;
}

=item date getLastClose(ledgerCode)

  This routine simply returns the last closeout date for this ledger.
  
=cut
sub getLastClose
{
  my $self = shift;
  my $ledgerCode = shift;
  if ($ledgerCode eq "")
  {
    $self->setError("Error! no ledgerCode sent to getLastClose!<br>\n");
    return undef;
  }
  my $sth = $self->{dbHandle}->read(sql => "SELECT MAX(bal_date) FROM acct_bal_tb WHERE ledger_code = '$ledgerCode'"); 
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError("Error in getLastClose!<br>\n" . $self->{dbHandle}->errorMessage);
    return undef;
  }
  if (my @info = $sth->fetchrow_array)
  {
    return $info[0];
  }
  return "";
}

=item % quickDate(range => 0|1, quarters => 0|1, months => 0|1, name => , target => , form => , ledgerCode => "", size => 4)

  This routine will create a date select box holding the dates given.  The javascript 
  output will take the date selected and set the value of the target to it.
  
  the returned hash will have {body}, {head} etc for parts of the html file
  
=cut
sub quickDate
{
  my $self = shift;
  my %args = (range => 0, quarters => 1, months => 0, name => "quickDate", target => "", form => "me.parent", ledgerCode => "", size => 4, @_);
  my $message = $self->{langObj}->map("selectDate");
  my %output = ('body' => "", 'head' => "" );
  if ($args{target} eq "")
  {  
    $self->prefixError("Invalid target sent to quickDate<br>");
    return undef; 
  }
  if ($args{ledgerCode} eq "")
  {  
    $self->prefixError("Invalid ledgerCode sent to quickDate<br>");
    return undef; 
  }
  my $ledger = $self->getValidLedger(ledgerCode => $args{ledgerCode});
  if (not defined $ledger)
  {  
    $self->prefixError("Invalid ledger object returned in quickDate<br>");
    return undef; 
  }
  
  $output{body} = <<"ENDOF";
  <select name="$args{name}" size="$args{size}" onChange="EnterDate();">
ENDOF
  my @dates = $ledger->getQuarterEnds($self->getCurrentFiscalYear(ledgerCode => $args{ledgerCode}));
  foreach my $date (@dates)
  {
    $output{body} .= <<"ENDOF";
      <option value="$date">$date</option>
ENDOF
  }
  $output{body} .= "</select>";
  $output{head} =  <<"ENDOF";
<script language="Javascript1.2">
  function EnterDate(var me)
  {
    var d = $args{form};
    if (d.$args{name}.selectedIndex == -1)
    {
      alert("$message");
      return;
    }
    d.$args{target}.value = d.$args{name}.options[d.$args{name}.selectedIndex].value;
  }
</script>
ENDOF
  return %output;
}
              
1;
__END__

=pod

  NOTE:  All data fields are accessible by specifying the object
         and pointing to the data member to be modified on the
         left-hand side of the assignment.
         Ex.  $obj->variable($newValue); or $value = $obj->variable;


=head1 AUTHOR

JT Moree, Xperience, Inc. <moreejt@pcxperience.org>

=head1 SEE ALSO

perl, Portal::Accounting

=cut

