# AccountObject.pm - The Object Class that provides a Account Object
# Created by James Pattie & Jason A. Pattie, 11/07/2000.
#$Id: AccountObject.pm,v 1.7 2003/07/23 19:18:06 pcxuser Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Objects::AccountObject;
use strict;
use Portal::Accounting::Objects::Base;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::Objects::Base Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '0.03';

my %trueFalse = ( 1 => "true", 0 => "false" );
my %falseTrue = ( "true" => 1, "false" => 0 );

=head1 NAME

AccountObject - Object used to build a Account Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::Objects::AccountObject;
  my $minLength = 5;  # This should come from the database.
  my $accountObj = Portal::Accounting::Objects::AccountObject->new(
            name => "test", number => "1234", type => 1,
            active => "true", taxable => "true");
  if ($accountObj->didErrorOccur())
  {
    die $accountObj->errorMessage();
  }

  # now you would pass the accountObj object to the createAccount
  # method in the Portal::Accounting::DB::Main module.

  # This object could also have been created by a getAccountInfo
  # or similar routine from the Portal::Accounting::DB::Main module.
  # In that case, creating = "false".

=head1 DESCRIPTION

AccountObject is a Account class.  This represents an account in the
accounting system.  Accounts are grouped by their type and must have
unique names and numbers.  The minAccountLength value should come from
the current configuration values that specify how long an account must
be when you create it.  The length of the number will be compared against
the minAccountLength to enforce this requirement, but only when creating
is true.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new(name, number, type, active, taxable)

 Creates a new instance of the Portal::Accounting::Objects::AccountObject module.
 See Portal::Base(3) for a listing of required arguments.

 requires:
  name
  number
  type
  active
  taxable

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( name => "", number => 0, type => -1, active => "true", taxable => "true", @_ );

  if ($self->didErrorOccur)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module
  $self->{name} = $args{name};
  $self->{number} = $args{number};
  $self->{type} = $args{type};
  $self->{active} = $args{active};
  $self->{taxable} = $args{taxable};

  # do validation
  if (!$self->isValid)
  {
    # the error is set in the isValid() method.
    return $self;
  }

  # do anything else you might need to do.

  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.
  if (length $self->{name} == 0)
  {
    $self->invalid("name", $self->{name});
  }
  if ($self->{type} < 0 || $self->{type} !~ /^\d+$/)
  {
    $self->invalid("type", $self->{type});
  }
  if ($self->{active} !~ /^(true|false)$/)
  {
    $self->invalid("active", $self->{active});
  }
  if ($self->{taxable} !~ /^(true|false)$/)
  {
    $self->invalid("taxable", $self->{taxable});
  }

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

=item string printXML(fullXML bool, indent int)

 This routine returns an XML snippet for this object for use in such
 things as exporting data

 fullXML - defaults to 0 (false).  If true, we output the xml header tag.
 indent - defaults to 2.  Specifies how many spaces in our top level tag
          needs to be.  If fullXML is true, then we force 0, otherwise we
          use what the user specifies.

=cut

sub printXML
{
  my $self = shift;
  my $result = "";
  my %args = ( fullXML => 0, indent => 2, @_ );
  my $fullXML = $args{fullXML};
  my $indent = ($fullXML ? 0 : $args{indent});  # start at 0 indentation if we are making a full XML document.
  my $spaceStr = " " x $indent;

  if ($self->isValid())
  {
    if ($fullXML)
    {
      $result .= "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>";
    }
    $result .= $spaceStr . "<account>\n";
    foreach my $name (qw{name number type taxable})
    {
      if ($name eq "number")
      {
        $result .= $spaceStr . "  <code>$self->{$name}</code>\n";
      }
      else
      {
        $result .= $spaceStr . "  <$name>$self->{$name}</$name>\n";
      }
    }
    $result .= $spaceStr . "</account>\n";
  }
  else
  {
    $result .= "Data not valid!\n\n";
    $result .= $self->errorMessage();
    die $result;
  }

  return $result;
}

=back

=cut

1;
__END__

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:admin@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::Base(3)

=cut
