# Base.pm - The Object Class that provides a Base Object
#$Id: Base.pm,v 1.5 2004/02/11 21:08:04 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2003 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Objects::Base;
use strict;
use Portal::Base;
use Portal::Data::Variables;
use Portal::Accounting::Data::Variables;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT %Elements);

require Exporter;

@ISA = qw(Portal::Base Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '0.01';
%Elements = ();

=head1 NAME

BaseObject - Object used to build a Base Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::Objects::Base;
  my $BaseObj = Portal::Accounting::Objects::Base->new(   );
  if ($BaseObj->didErrorOccur())
  {
    die $BaseObj->errorMessage();
  }

=head1 DESCRIPTION

BaseObject is a Base class for Accounting Objects such as Objects::Ledger and Objects::Account.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new( )

    Creates a new instance of Portal::Accounting::Objects::Base.
    See Portal::Base(3) for a listing of required arguments.

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );

  if ($self->didErrorOccur)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module
  eval { $self->{portalVariables} = Portal::Data::Variables->new(langObj => $self->{langObj}); };
  if ($@)
  {
    $self->setError($@);
    return $self;
  }
  if ($self->{portalVariables}->didErrorOccur)
  {
    $self->setError($self->{portalVariables}->errorMessage);
    return $self;
  }
  eval { $self->{variables} = Portal::Accounting::Data::Variables->new(langObj => $self->{langObj}); };
  if ($@)
  {
    $self->setError($@);
    return $self;
  }
  if ($self->{variables}->didErrorOccur)
  {
    $self->setError($self->{variables}->errorMessage);
    return $self;
  }

  # do validation
  if (!$self->Portal::Accounting::Objects::Base::isValid)
  {
    # the error is set in the isValid() method.
    return $self;
  }

  # do anything else you might need to do.
  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.
  if (not defined $self->{portalVariables})
  {
    $self->setError("portalVariables is not defined!<br>\n");
    return 0;
  }

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

=item string print()

  This routine returns a text string that represents the data in this object.

=cut

sub print
{
  return "This is the Objects::Base print method.\n"
}

=item string printXML(fullXML bool, indent int)

 This routine returns an XML snippet for this object for use in such
 things as exporting data

 fullXML - defaults to 0 (false).  If true, we output the xml header tag.
 indent - defaults to 2.  Specifies how many spaces in our top level tag
          needs to be.

=cut

=item string printHTML()

  This routine returns a string of html that represents this object.

=cut

=item string printForm(prefix => "")

  This routine returns xhtml form elements that can be used to edit this object.
  A prefix can be used to uniquely identify this form from others of the same 
  type.  The names of form elements can be obtained in a hash:

  use Portal::Accounting::Objects::Base;
  my %elements = %Portal::Accounting::Objects::Base::Elements;

  !!Make a copy of the hash as shown above--NOT a refence.
  You would generally place your input from the user within the Elements hash.
  Then you can pass the Elements hash to the new method
  of a Base object instead of individual arguments.

  foreach my $prefix (%prefixes)
    foreach $index (key %Elements)
      $tempHash{$index} = $input->{$prefix . $index}
    my object = Portal::Accounting::Objects::Base->new(%tempHash);

=cut

=back

=cut

1;
__END__

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:xiwa@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal::Base(3)

=cut
