# FiscalYearObject.pm - The Object Class that provides a FiscalYear Object
# Created by James Pattie & Jason A. Pattie, 11/07/2000.
#$Id: FiscalYearObject.pm,v 1.2 2004/01/08 18:06:28 pcxuser Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Objects::FiscalYearObject;
use strict;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Exporter AutoLoader);
@EXPORT = qw();
	
$VERSION = '0.01';

# new 
# instantiates an instance of a FiscalYearObject
# requires: year, closed, startDate, endDate, q1Start, q1End, q2Start, q2End, q3Start, q3End, q4Start, q4End, ledgerCode
# optional:
# returns: FiscalYearObject
sub new
{
  my $that = shift;
  my $class = ref($that) || $that;
  my $self = bless {}, $class;
  my %args = ( year => "2001", closed => "false", startDate => "", endDate => "", q1Start => "", q1End => "", q2Start => "", q2End => "",
               q3Start => "", q3End => "", q4Start => "", q4End => "", ledgerCode => "", @_ );
  my $errStr = "FiscalYearObject: new() - Error!<br>\n";
               
  $self->{year} = $args{year};
  $self->{closed} = $args{closed};
  $self->{startDate} = $args{startDate};
  $self->{endDate} = $args{endDate};
  $self->{q1Start} = $args{q1Start};
  $self->{q1End} = $args{q1End};
  $self->{q2Start} = $args{q2Start};
  $self->{q2End} = $args{q2End};
  $self->{q3Start} = $args{q3Start};
  $self->{q3End} = $args{q3End};
  $self->{q4Start} = $args{q4Start};
  $self->{q4End} = $args{q4End};
  $self->{ledgerCode} = $args{ledgerCode};

  # do validation
  $self->{error} = !$self->isValid;  
  if ($self->{error})
  {
    $self->{errorString} = $errStr . $self->{errorString};
  }

  return $self;
}

# isValid - Returns 0 or 1 to indicate if the object is valid.
sub isValid
{
  my $self = shift;
  my $error = 0;
  my $errorString = "";
  my $errStr = "FiscalYearObject: isValid() - Error!<br>\n";

  # do validation code here.
  
  if ($self->{year} !~ /^\d{4}$/)
  {
    $error = 1;
    $errorString .= "year = '$self->{year}' is invalid!<br>\n";
  }
  if ($self->{closed} !~ /^(true|false)$/)
  {
    $error = 1;
    $errorString .= "closed = '$self->{closed}' is invalid!<br>\n";
  }
  if (length $self->{ledgerCode} == 0)
  {
    $error = 1;
    $errorString .= "ledgerCode = '$self->{ledgerCode}' is invalid!<br>\n";
  }
  foreach my $date (qw(q1Start q1End q2Start q2End q3Start q3End q4Start q4End))
  {
    if ($self->{$date} !~ /^(\d{4}(-|\/)\d{2}(-|\/)\d{2}|\d{2}(-|\/)\d{2}(-|\/)\d{4})$/)
    {
      $error = 1;
      $errorString .= "$date = '$self->{$date}' is invalid!<br>\n";
    }
  }

  $self->{error} = $error;
  $self->{errorString} = $errStr if $error;
  $self->{errorString} .= $errorString;

  return !$error;
}

sub DESTROY
{
  my $self = shift;
}

sub AUTOLOAD
{
  my $self = shift;
  my $type = ref($self) || die "$self is not an object";
  my $name = $AUTOLOAD;
  $name =~ s/.*://;	# strip fully-qualified portion
  unless (exists $self->{$name})
  {
    die "Can't access `$name' field in object of class $type";
  }
  if (@_)
  {
    return $self->{$name} = shift;
  }
  else
  {
    return $self->{$name};
  }
}

# setError
# requires: errorString
# returns: nothing
sub setError
{
  my $self = shift;
  my %args = ( errorString => "", @_ );
  my $errorString = $args{errorString};
  
  $self->{errorString} = $errorString;
  $self->{error} = 1;
}

# didErrorOccur
# returns the value of error.
sub didErrorOccur
{
  my $self = shift;

  return $self->{error};
}

# errorMessage
# returns the value of errorString
sub errorMessage
{
  my $self = shift;

  return $self->{errorString};
}

1;
__END__

=head1 NAME

FiscalYearObject - Object used to build a FiscalYear Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::Objects::FiscalYearObject;
  my $fyObj = Portal::Accounting::Objects::FiscalYearObject->new(
              year => "2000", closed => "false", 
              startDate => "01/01/2000", endDate => "12/31/2000", 
              q1Start => "01/01/2000", q1End => "03/31/2000", 
              q2Start => "04/01/2000", q2End => "06/30/2000", 
              q3Start => "07/01/2000", q3End => "09/30/2000", 
              q4Start => "10/01/2000", q4End => "12/31/2000", 
              legerCode => "test");
  if ($fyObj->didErrorOccur())
  {
    die $fyObj->errorMessage();
  }

=head1 DESCRIPTION

FiscalYearObject is a FiscalYear class.  This represents an entry
in the known_years_tb table in the database.

=head1 Exported FUNCTIONS

  scalar new(year, closed, startDate, endDate, q1Start, q1End, q2Start, 
             q2End, q3Start, q3End, q4Start, q4End, ledgerCode)
    Creates a new instance of the
    Portal::Accounting::Objects::FiscalYearObject object.
    
    year is a 4 digit value format YYYY.
    closed is a boolean.
    ledgerCode is a string.
    all other values are dates in MM/DD/YYYY format.

  bool isValid(void)
    Returns 0 or 1 to indicate if the object is valid.  The error will
    be available via errorMessage().

  void setError(errorString)
    Sets error = 1 and $self->{errorString} = errorString.

  scalar didErrorOccur(void)
    Returns the value of error.

  scalar errorMessage(void)
    Returns the value of errorString.


  NOTE:  All data fields are accessible by specifying the object
         and pointing to the data member to be modified on the
         left-hand side of the assignment.
         Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

PC & Web Xperience, Inc. (mailto:admin@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3)

=cut
