# Object that contains the referenceNum, ledgerCode, journalCode, transDate, payeePayer, memo, remember
# and an array of JournalEntryRows.
#$Id: JournalEntry.pm,v 1.3 2004/02/11 22:02:31 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Objects::JournalEntry;

use strict;
use Portal::Accounting::Objects::JournalEntryRow;
use vars qw($VERSION @ISA @EXPORT $AUTOLOAD);

require Exporter;

@ISA = qw(Portal::Accounting::Objects::Base Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.3 $';

# new
# instantiates an instance of a JournalEntry
sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);

  $self->{referenceNum} = "-1";
  $self->{ledgerCode}   = "G";
  $self->{journalCode}  = "";
  $self->{transDate}    = "";
  $self->{payeePayer}   = "";
  $self->{memo}         = "";
  $self->{remember}     = "false";
  $self->{rowObjects}   = [];
  $self->{trackingId}   = "";
  $self->{origId}       = "";

  return $self;
}

sub populate
{
  my $self = shift;
  my %args = ( referenceNum => '-1',
      ledgerCode => undef,
      journalCode => 'G',
      transDate => undef,
      payeePayer => '',
      memo => '',
      remember => 'false',
      rowObjects => [],
      trackingId => "",
      origId => "",
      input => undef,
      @_ );

  $self->{referenceNum} = $args{referenceNum};
  $self->{ledgerCode}   = $args{'ledgerCode'};
  $self->{journalCode}  = $args{'journalCode'};
  $self->{transDate}    = $args{'transDate'};
  $self->{payeePayer}   = $args{'payeePayer'};
  $self->{memo}         = $args{'memo'};
  $self->{remember}     = $args{'remember'};
  $self->{rowObjects}   = $args{'rowObjects'};
  $self->{trackingId}   = $args{trackingId};
  $self->{origId}       = $args{origId};

  return $self->isValid;
}

# isValid
# validates the object in its current state
sub isValid
{
  my $self = shift;
  # Can we call the parent at the bottom to pick up on the error here without repeating code
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  my $errorHeader = "JournalEntry->isValid(): ";

  if ($self->{referenceNum} = -1 || $self->{referenceNum} !~ /^\d\d*/)
  {
    $self->invalid(name => "referenceNum", value => $self->{referenceNum});
  }
  if ($self->{ledgerCode} eq "")
  {
    $self->missing(name => "ledgerCode");
  }
  if (not exists $self->{variables}->{journalCodes}->{$self->{journalCode}})
  {
    $self->invalid(name => "journalCode", value => $self->{journalCode});
  }
  if ($self->{transDate} =~ /^\d{8}$/)
  {
    $self->invalid(name => "transDate", value => $self->{transDate});
  }
  if (($self->{remember} ne "true" and $self->{remember} ne "false"))
  {
    $self->{remember} = "false";
  }
  if (scalar @{ $self->{rowObjects} } < 2)
  {
    $self->prefixError("The JournalEntry rows array contains too few items (" . (scalar @{ $self->{rowObjects} }) . ")!");
  }
  my $sum = 0;
  my $debit_sum = 0;
  my $credit_sum = 0;
  for (my $rowcount=0; $rowcount < scalar @{ $self->{rowObjects} }; $rowcount++)
  {
    if (ref($self->{rowObjects}[$rowcount]) ne "Portal::Accounting::Objects::JournalEntryRow")
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Row $rowcount is not a valid JournalEntryRow (" . ref($self->{rowObjects}[$rowcount]) . ")!";
    }
    if (!$self->{rowObjects}[$rowcount]->isValid())
    {
      $self->{error} = 1;
      $self->{errorString} = "JournalEntry->JournalEntryRow->isValid(): Row $rowcount has invalid data! " . $self->{rowObjects}[$rowcount]->errorMessage;
    }
    else
    {
      if ($self->{rowObjects}[$rowcount]->dorc eq "true")
      {
        $debit_sum += $self->{rowObjects}[$rowcount]->amount;
      }
      else
      {
        $credit_sum += $self->{rowObjects}[$rowcount]->amount;
      }
    }
  }
  $sum = sprintf("%.2f", $debit_sum - $credit_sum);
  if ($sum != 0.00)
  {
    if ($debit_sum == 0.00)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "No debits entered!<br>Debit Sum: " . $debit_sum . "<br>Credit Sum: " . $credit_sum . "<br>Sum: " . $sum;
    }
    elsif ($credit_sum == 0.00)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "No credits entered!<br>Debit Sum: " . $debit_sum . "<br>Credit Sum: " . $credit_sum . "<br>Sum: " . $sum;
    }
    if ($sum < 0.00)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Credits are greater than debits!<br>Debit Sum: " . $debit_sum . "<br>Credit Sum: " . $credit_sum . "<br>Sum: " . $sum;
    }
    elsif ($sum > 0.00)
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Debits are greater than credits!<br>Debit Sum: " . $debit_sum . "<br>Credit Sum: " . $credit_sum . "<br>Sum: " . $sum;
    }
  }
  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

# addRow
# takes the values for a JournalEntryRow and creates a new object to add to the array of objects in the
# JournalEntry
# returns newly added row number on success, 0 on failure
sub addRow
{
  my $self = shift;
  my $errorHeader = "JournalEntry->addRow(): ";
  my %args = ( @_ );

  my $rowDate       = $args{'rowDate'};
  my $comment       = $args{'comment'};
  my $subEntity     = $args{'subEntity'};
  my $accountNum    = $args{'accountNum'};
  my $subAccountNum = $args{'subAccountNum'};
  my $dorc          = $args{'dorc'};
  my $amount        = $args{'amount'};
  my $cleared       = $args{'cleared'};

  if (not exists $args{'rowDate'})
  {
    $rowDate = $self->{transDate};
  }
  if (not exists $args{'comment'})
  {
    $comment = "";
  }
  if (not exists $args{'subEntity'})
  {
    $subEntity = "";
  }
  if (not exists $args{'accountNum'})
  {
    $self->{error} = 1;
    $self->{errorString} = "JournalEntry->addRow(): accountNum value missing!";
    return 0;
  }
  if (not exists $args{'subAccountNum'})
  {
    $subAccountNum = -1;
  }
  if (not exists $args{'dorc'})
  {
    $self->{error} = 1;
    $self->{errorString} = "JournalEntry->addRow(): dorc value missing!";
    return 0;
  }
  if (not exists $args{'amount'})
  {
    $self->{error} = 1;
    $self->{errorString} = "JournalEntry->addRow(): amount value missing!";
    return 0;
  }
  if (not exists $args{'cleared'})
  {
    $cleared = ' ';
  }

  # create a new JournalEntryRow with the values passed in
  my $transrowobj = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => "$rowDate", comment => "$comment", subEntity => "$subEntity", accountNum => $accountNum, subAccountNum => $subAccountNum, dorc => "$dorc", amount => "$amount", cleared => "$cleared");
  if ($transrowobj->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Failed to add row!<br>\n" . $transrowobj->errorMessage;
    return 0;
  }
  # add the row object to the current array of row objects
  $self->{rowObjects}[scalar @{$self->{rowObjects}}] = $transrowobj;

  return (scalar @{$self->{rowObjects}} - 1);
}

# deleteRow
# Removes a row from the rowObjects array
# returns 1 on success, 0 on failure
sub deleteRow
{
  my $self = shift;
  my $errorHeader = "JournalEntry->deleteRow(): ";
  my %args = ( @_ );
  my $row = $args{'row'};

  if (not exists $args{'row'})
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Row number value argument missing!";
    return 0;
  }
  if ($row !~ /^-?[0-9]+$/)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Invalid row number argument!  The value does not appear to be numeric.";
    return 0;
  }

  if ($row > 0 && $row < scalar @{$self->{rowObjects}})
  {
    # remove the row from the array using splice to "cut" the element out
    splice(@{$self->{rowObjects}}, $row, 1);
  }

  return 1;
}

1;
__END__

=head1 NAME

JournalEntry - Object used to build a complete JournalEntry for storage in the accounting database.

=head1 SYNOPSIS

  use JournalEntry;
  my $transobj = JournalEntry->new(referenceNum => '379',
  					ledgerCode => 'ts1',
  					journalCode => 'G',
  					transDate => '12-14-1999',
  					payeePayer => 'Bank',
  					memo => 'cashed check',
  					remember => 'false',
  					rowObjects => \@RowObjectArray,
            origId => "");
  if ($transobj->didErrorOccur())
  {
    die $transobj->errorMessage();
  }

  # set object data by specifying the data field desired
  $transobj->journalCode("I");
  $transobj->memo("");

  # make sure the object is useful by calling the validation routine
  # before writing it to the database
  if ($transobj->isValid())
  {
    updateJournalEntry($transobj);
  }
  else
  {
    die $transobj->errorMessage();
  }

=head1 DESCRIPTION

JournalEntry is a class that stores data for an accounting
JournalEntry.  It provides a container for a JournalEntry that will be used
to write to the database or update an existing record in the database.

=head1 Exported FUNCTIONS

  scalar new(referenceNum, ledgerCode, journalCode, transDate,
             payeePayer, memo, remember, rowObjects, trackingId, origId)
    Creates a new instance of the Portal::Accounting::Objects::JournalEntry
    object.  A new instance can be created in an invalid state (i.e., missing
    fields, etc.).  isValid() must be called on the object before using the
    object in a database JournalEntry.

    trackingId = the genSysId value from an XML JournalEntry Import. Can be "".
    origId = the original referenceNum value for this entry.  Can be "".

  void setError(errorString)
    Sets error = 1 and $self->{errorString} = errorString.

  scalar didErrorOccur(void)
    Returns the value of error.

  scalar errorMessage(void)
    Returns the value of errorString.

  scalar isValid(void)
    Runs through the data fields contained in the class to make sure
    they contain valid data.  This function should be called before
    using this object in a database JournalEntry.

  scalar addRow(rowDate, comment, subEntity, accountNum, subAccountNum,
                dorc, amount, cleared)
    Adds a row to the rowObjects array in the JournalEntry.
    Returns the row number of the newly added row or 0 on failure.

  scalar deleteRow(row)
    Removes the row specified from the rowObjects array.

  NOTE:  All data fields are accessible by specifying the object
         and pointing to the data member to be modified on the
         left-hand side of the assignment.

=head1 AUTHOR

James A. Pattie (mailto:james@pcxperience.com), Jason A. Pattie (mailto:pattieja@pcxperience.com)

=head1 SEE ALSO

perl(1), JournalEntry(3), DBIWrapper(3)

=cut
