# Ledger.pm - The Object Class that provides a Ledger Object
# Created by James Pattie & Jason A. Pattie, 11/07/2000.
#$Id: Ledger.pm,v 1.17 2004/08/13 22:05:40 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Objects::Ledger;
use strict;
use Portal::Accounting::Objects::Base;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::Objects::Base Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '0.04';

=head1 NAME

Ledger - Object used to build a Ledger Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::Objects::Ledger;
  my $ledgerObj = Portal::Accounting::Objects::Ledger->new(
            name => "Testing", code => "test", color => "blue",
            cashOrAccrual => "true", businessType => 2, fyMo => 12,
            fyDay => 31, fyCurrent => "true");
  if ($ledgerObj->didErrorOccur())
  {
    die $ledgerObj->errorMessage();
  }

=head1 DESCRIPTION

Ledger is a Ledger class.  This represents an entry in the
ledger_tb table in the database.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new(name, code, color, cashOrAccrual, businessType, )

    Creates a new instance of Portal::Accounting::Objects::Ledger.
    See Portal::Base(3) for a listing of required arguments.

    name      - text
    code      - text
    color     - text
    cashOrAccrual - string (cash, accrual)
    businessType  - integer pertaining to a business type (see business_type_tb)
    businessTypeName  - a business type name from the business_type_tb that will be dereferenced into a language specific string
    knownYears- a hash containing fiscal years as keys (2000, 2001) and the ending date for the year as data.
        examples.
        $knownYears->{2000} = '2001-01-31'
        $knownYears->{2000} = '20001231'
    start and offset   - array of month beginnings in format MM-DD or MMDD. Starting with index 1 for 1st month, etc.
        start[0]           - offset, if the fiscal year starts in the previous year this should be -1, otherwise 0. i.e. FY 2000 starts 1999-12-01
        start[0]  = "-1"
        start[1]  = "MMDD"

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = (
    name => "",
    code => "",
    color => "",
    cashOrAccrual => "",
    businessType => "",
    businessTypeName => "",
    knownYears => undef,
    start => undef,
     @_ );

  if ($self->didErrorOccur)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module
  $self->{name} = $args{name};
  $self->{code} = $args{code};
  $self->{color} = $args{color};
  $self->{cashOrAccrual} = $args{cashOrAccrual};
  $self->{businessType} = $args{businessType};
  $self->{businessTypeName} = $args{businessTypeName};
  $self->{start} = $args{start};
  $self->{knownYears} = $args{knownYears};

  # do validation
  if (!$self->isValid)
  {
    # the error is set in the isValid() method.
    return $self;
  }

  # do anything else you might need to do.
  $self->{offsetCache} = ();
  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.

  if (length $self->{name} == 0)
  {
    $self->missing("name");
  }
  if (length $self->{code} == 0)
  {
    $self->missing("code");
  }
  if ($self->{cashOrAccrual} eq "")
  {
    $self->missing("cashOrAccrual");
  }
  elsif ($self->{cashOrAccrual} !~ /^(cash|accrual)$/i)
  {
    $self->invalid("cashOrAccrual", $self->{cashOrAccrual});
  }
  if (not defined $self->{knownYears})
  {
    $self->missing("knownYears");
  }
  else
  {
    foreach my $year (keys %{$self->{knownYears}} )
    {
      if ($year !~ /^(\d{4})$/) #check the key for a valid integer (year)
      {
        $self->invalid('knownYears_'.$year, $year);
      }
      elsif ($self->{knownYears}->{$year} !~ m/^(\d{8})$/ && $self->{knownYears}->{$year} !~ s/^(\d{4})([-\/\\])(\d{1,2})([-\/\\])(\d{1,2})$/$1.sprintf("%02d", $3 ).sprintf("%02d",$5)/e)
      {
        $self->invalid("knownYears->$year", $self->{knownYears}->{$year});
      }
    }
  }
  if ($self->{businessTypeName} eq "")
  {
    $self->missing("businessTypeName");
  }
  if ($self->{businessType} eq "")
  {
    $self->missing("businessType");
  }
  elsif ($self->{businessType} !~ /^(\d)$/)
  {
    $self->invalid("businessType", $self->{businessType});
  }

  if ($self->{start}[0] !~ m/^(-?\d+)$/)
  {
    $self->invalid('start[0]/offset', $self->{start}[0]);
  }

  my $years = 0;
  for (my $index = 1; $index < 13; $index++)
  {
    if ($self->{start}[$index] =~ m/^(\d{4})$/ || $self->{start}[$index] =~ s/^(\d{1,2})([^\d])(\d{1,2})$/sprintf("%02d",$1).sprintf("%02d",$3)/e)
    {
      #skip month1
      if ($index > 1)
      {
        #compare this date+offset with the last one+offset
        if ($self->{start}[$index] < $self->{start}[$index - 1])
        {
          $years++;
          if ( $years > 1 )
          {
            $self->invalid("start[".$index."]", $self->{start}[$index] . ": is out of order");
            $years = 1;
          }
        }
      }
    }
    else
    {
      $self->invalid("start[".$index."]", $self->{start}[$index]);
    }
  }

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

# =item String[] businessType(void)
#
#  Returns the business type number
#
# =cut
#
# sub businessType
# {
#   my $self = shift;
#   return $self->{businessType};
# }

=item String[] businessTypeName(void)

 Returns the business type name from the proper lanquage
 if it is set in the object or an empty string if not set.
 There is no guarantee that it is set!

=cut

sub businessTypeName
{
  my $self = shift;
  if ($self->{businessTypeName} ne "")
  {
    return $self->{langObj}->map($self->{businessTypeName});
  }

  return "";
}

=item date[] getQuarterEnds([fiscal year])

 Returns an array of strings representing the ending dates of Quarters
 for this ledger.  If a year is given, then the dates will contain
 the year and be in the format YYYY-MM-DD, otherwise the format
 will be MM-DD.

=cut

sub getQuarterEnds
{
  my $self = shift;
  my @quarters = ();
  my $year;
  my $month = 4;
  my $strip = "n";
  if (@_ == 0)
  {
    $year = 2000;
    $strip = "y";
  }
  elsif (scalar @_ == 1)
  {
    $year = @_[0];
  }
  if ($year !~ m/^(\d)+$/)
  { return undef; }

  for (my $index = 0; $index < 4; $index++)
  {
    if ($self->{start}[$month] !~ m/^(\d{2})(\d{2})$/)
    {
      $self->setError("Bad internal quarter start: " . $self->{start}[$month]);
      return undef;
    }
    #date:calc subtract 1 day from next quarter start date
    my @tempdate = Date::Calc::Add_Delta_Days($year + $self->offset($month) , $1, $2, -1);
    if ($strip eq "y")
    {
      @quarters[$index] = sprintf("%02d",$tempdate[1])."-".sprintf("%02d",$tempdate[2]);
    }
    else
    {
      @quarters[$index] = $tempdate[0]."-".sprintf("%02d",$tempdate[1])."-".sprintf("%02d",$tempdate[2]);
    }
    $month += 3;
    if ($month == 13)  #wraps around to the next year.
    {
      $month = 1;
      $year++;
    }
  }

  return @quarters;
}

=item date[] getQuarterStarts([fiscal year])

 Returns an array of strings representing the starting dates of Quarters
 for this ledger.  If a year is given, then the dates will contain
 the year and be in the format YYYY-MM-DD, otherwise the format
 will be MM-DD.

=cut

sub getQuarterStarts
{
  my $self = shift;
  my @quarters = ();
  my $year;
  my $month = 4;
  my $strip = "n";
  if (@_ == 0)
  {
    $strip = "y";
  }
  elsif (@_ == 1)
  {
    $year = @_[0];
    if ($year !~ m/^(\d)+$/)
    { return undef; }
  }

  for (my $index = 1; $index < 5; $index++)
  {
    if ($self->{start}[$month] !~ m/^(\d{2})(\d{2})$/)
    {
      $self->setError("Bad internal quarter start: " . $self->{start}[$index]);
      return undef;
    }
    my $tempIndex;
    if ($index == 4)
    {  $tempIndex = 0; }
    else
    {  $tempIndex = $index; }
    if ($strip eq "y")
    {
      @quarters[$tempIndex] = sprintf("%02d",$1)."-".sprintf("%02d",$2);
    }
    else
    {
      @quarters[$tempIndex] = ($year + $self->offset($month))."-".sprintf("%02d",$1)."-".sprintf("%02d",$2);
    }
    $month += 3;
    if ($month == 13)
    {  $month = 1; }
  }

  return @quarters;
}

=item date[] getMonthEnds([fiscal year])

 Returns an array of strings representing the ending dates of months
 for this ledger.  If a year is given, then the dates will contain
 the year and be in the format YYYY-MM-DD, otherwise the format
 will be MM-DD.

=cut

sub getMonthEnds
{
  my $self = shift;
  my @months = ();
  my $year;
  my $strip = "n";
  if (scalar @_ == 0)
  {
    $year = 2000;
    $strip = "y";
  }
  elsif (scalar @_ == 1)
  {
    $year = @_[0];
  }
  if ($year !~ m/^(\d)+$/)
  { return undef; }

  for (my $index = 1; $index <= 12; $index++)
  {
    if ($self->{start}[$index] !~ m/^(\d{2})(\d{2})$/)
    {
      $self->setError("Bad internal month start: " . $self->{start}[$index]);
      return undef;
    }
    #date:calc subtract 1 day from next month start date
    #if on 0 index which is the first month, add 1 to the year so that we get the last day of the current year instead of last year, then put in the last spot [11]
    my $tempYear = $year;
    my $tempIndex = $index;
    if ($index == 1)
    {
      $tempYear++;
      $tempIndex = 12;
    }
    my @tempdate = Add_Delta_Days($tempYear + $self->offset($index) , $1, $2, -1);
    if ($strip eq "y")
    {
      @months[$tempIndex-1] = sprintf("%02d",$tempdate[1])."-".sprintf("%02d",$tempdate[2]);
    }
    else
    {
      @months[$tempIndex-1] = $tempdate[0]."-".sprintf("%02d",$tempdate[1])."-".sprintf("%02d",$tempdate[2]);
    }
  }

  return @months;
}

=item date[] getMonthStarts([fiscal year])

 Returns an array of strings representing the starting dates of months
 for this ledger.  If a year is given, then the dates will contain
 the year and be in the format YYYY-MM-DD, otherwise the format
 will be MM-DD.

=cut

sub getMonthStarts
{
  my $self = shift;
  my @months = ();
  my $year;
  my $strip = "n";
  if (scalar @_ == 0)
  {
    $strip = "y";
  }
  elsif (scalar @_ == 1)
  {
    $year = @_[0];
    if ($year !~ m/^(\d)+$/)
    { return undef; }
  }

  for (my $index = 1; $index <= 12; $index++)
  {
    if ($self->{start}[$index] !~ m/^(\d{2})(\d{2})$/)
    {
      $self->setError("Bad internal month start: " . $self->{start}[$index]);
      return undef;
    }
    if ($strip eq "y")
    {
      @months[$index-1] = $1."-".$2;
    }
    else
    {
      @months[$index-1] = ($year + $self->offset($index))."-".sprintf("%02d",$1)."-".sprintf("%02d",$2);
    }
  }

  return @months;
}

=item int offset(int)

 This routine calculates the offset for each month to change the
 fiscal year value into a calendar year value.

=cut

sub offset
{
  my $self = shift;
  my $month = shift;

  if ($month !~ m/^(\d)+$/ || $month < 1 || $month > 12)
  { return undef } ;
  if (not defined $self->{offsetCache}[$month])
  {
    $self->{offsetCache}[1] = $self->{start}[0];
    my $offset = $self->{start}[0];
    for (my $index = 2; $index < 13; $index++)
    {
      #compare this date+offset with the last one+offset
      if ($self->{start}[$index] < $self->{start}[$index - 1])
      {
        $offset++;
      }
      $self->{offsetCache}[$index] = $offset;
    }
  }
  return $self->{offsetCache}[$month];
}

=item string start(month => , fiscalYear =>) or start(<month>)

 This routine returns the starting date of the month specified.
 If a year is given, it will be translated to actual (YYYY-MM-DD).
 Otherwise format will be MM-DD.

=cut

sub start
{
  my $self = shift;
  my $month;
  my $year = "";
  my %args;
  my $return = "";
  if (@_ == 1)
  {
    $month = @_[0];
  }
  else
  {
    %args = (month => "", fiscalYear => "", @_);
    $month = $args{month};
    $year = $args{fiscalYear};
  }
  if ($month !~ m/^(\d)+$/ || $month < 1 || $month > 12)
  {
     $self->prefixError($self->{langObj}->map("Invalid") . " " . $self->{langObj}->map("month") . " '". $month . "'<br>\n");
     return undef;
  }

  if (defined $self->{start}[$month])
  {
    if ($self->{start}[$month] !~ m/^(\d{2})(\d{2})$/)
    {
      $self->prefixError("Invalid internal month start: '".$self->{start}[$month]."'<br>\n");
      return undef;
    }
    $return = $1."-".$2;
    if ($year eq "")
    {  return $return; }
    elsif ($year =~ m/^\d+$/)
    {  return $year."-".$return;  }
    $self->prefixError("Invalid year: '".$year."'<br>\n");
  }

  $self->prefixError("Error in Objects::Ledger->start");
  return undef;
}

=item string end(month => , fiscalYear =>) or end(<month>)

 This routine returns the ending date of the month specified.
 If a year is given, it will be translated to actual (YYYY-MM-DD).
 Otherwise format will be MM-DD.

=cut
sub end
{
  my $self = shift;
  my $month;
  my $year = "";
  my %args;
  my $return = "";
  if (@_ == 1)
  {
    $month = @_[0];
  }
  else
  {
    %args = (month => "", fiscalYear => "", @_);
    $month = $args{month};
    $year = $args{fiscalYear};
  }
  if ($month !~ m/^(\d)+$/ || $month < 1 || $month > 12)
  {
     $self->prefixError($self->{langObj}->map("Invalid") . " " . $self->{langObj}->map("month") . " '". $month . "'<br>\n");
     return undef;
  }

  if (defined $self->{end}[$month])
  {
    if ($self->{end}[$month] !~ m/^(\d{2})(\d{2})$/)
    {
      $self->prefixError("Invalid internal month end: '".$self->{end}[$month]."'<br>\n");
      return undef;
    }
    $return = $1."-".$2;
    if ($year eq "")
    {  return $return; }
    elsif ($year =~ m/^\d+$/)
    {  return $year."-".$return;  }
    $self->prefixError("Invalid year: '".$year."'<br>\n");
  }

  $self->prefixError("Error in Objects::Ledger->end");
  return undef;
}

=item ISO-Date firstDate()

 Returns the first date the books were opened for.

=cut

sub firstDate
{
  my $self = shift;

  # try not to calculate it every time.
  if (!defined $self->{firstDate})
  {
    my @years = keys %{$self->{knownYears}};
    (my $start1 = $self->{start}->[1]) =~ s/^(\d{2})(\d{2})$/$1-$2/;

    # we build this up based upon the oldest known year + offset of first month and
    # tack on the first month-day entry.
    my $firstDate = ((sort { $a <=> $b } @years)[0] + $self->offset(1)) . "-" . $start1;

    $self->{firstDate} = $firstDate;
  }

  return $self->{firstDate};
}

=item ISO-Date formatDate(string)

 This routine returns the input date formatted in the ISO standard
 if a year is specified.  In either case, we have -'s for seperators.

 If an invalid date is specified, we return "" and set an error.

=cut

sub formatDate
{
  my $self = shift;
  my $date = shift;

  if ($date =~ /^(\d{4})(\d{2})(\d{2})$/)
  {
    $date = "$1-$2-$3";
  }
  elsif ($date =~ /^(\d{2})(\d{2})$/)
  {
    $date = "$1-$2";
  }
  elsif ($date !~ /^((\d{4}-)?\d{2}-\d{2})$/)
  {
    $self->invalid("date", $date);
    $date = "";
    $self->postfixError($self->genErrorString("all"));
  }

  return $date;
}

=item string printXML(fullXML bool, indent int)

 This routine returns an XML snippet for this object for use in such
 things as exporting data

 fullXML - defaults to 0 (false).  If true, we output the xml header tag.
 indent - defaults to 2.  Specifies how many spaces in our top level tag
          needs to be.  If fullXML is true, then we force 0, otherwise we
          use what the user specifies.

=cut

sub printXML
{
  my $self = shift;
  my $result = "";
  my %args = ( fullXML => 0, indent => 2, @_ );
  my $fullXML = $args{fullXML};
  my $indent = ($fullXML ? 0 : $args{indent});  # start at 0 indentation if we are making a full XML document.
  my $spaceStr = " " x $indent;

  if ($self->isValid())
  {
    if ($fullXML)
    {
      $result .= "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>";
      $result .= $spaceStr . "<ledger name=\"$self->{ledger}->{name}\" code=\"$self->{ledger}->{code}\">\n";
      $indent += 2;
      $spaceStr = " " x $indent;
    }
    $result .= $spaceStr . "<ledgerInfo>\n";
    foreach my $name (qw{color cashOrAccrual businessType businessTypeName})
    {
      $result .= $spaceStr . "  <$name>" . ($name eq "firstDate" ? $self->formatDate($self->{$name}) : ($name eq "businessTypeName" ? $self->businessTypeName() : $self->{$name})) . "</$name>\n";
    }
    $result .= $spaceStr . "  <knownYears>\n";
    foreach my $year (keys %{$self->{knownYears}})
    {
      $result .= $spaceStr . "    <info year=\"$year\" endDate=\"" . $self->formatDate($self->{knownYears}->{$year}) . "\"/>\n";
    }
    $result .= $spaceStr . "  </knownYears>\n";
    $result .= $spaceStr . "  <dates>\n";
    for (my $i=1; $i <= 12; $i++)
    {
      $result .= $spaceStr . "    <info index=\"$i\" start=\"" . $self->start(month => $i) . "\" offset=\"" . $self->offset($i) . "\"/>\n";
    }
    $result .= $spaceStr . "  </dates>\n";
    $result .= $spaceStr . "</ledgerInfo>\n";
    if ($fullXML)
    {
      $indent -= 2;
      $spaceStr = " " x $indent;
      $result .= $spaceStr . "</ledger>\n";
    }
  }
  else
  {
    $result .= "Data not valid!\n\n";
    $result .= $self->errorMessage();
    die $result;
  }

  return $result;
}

=back

=cut

1;
__END__

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:xiwa@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal::Base(3)

=cut
